"""
Describing of parent class ItemN.

It should be used ass parent to describe new item classes.
"""
import os
import os.path
from typing import NoReturn, Union
from pycifstar import Data, to_data

from cryspy.A_functions_base.function_1_markdown import md_to_html
from cryspy.A_functions_base.function_1_strings import find_prefix, \
    string_to_value_error_mark, value_error_mark_to_string
from cryspy.A_functions_base.function_1_objects import get_functions_of_objet


def string_to_fixed_or_refined(s_value: str):
    flag_refined = False
    try:
        value = float(s_value.split("(")[0])
        flag = (("(" in s_value) and (")" in s_value))
        if flag:
            flag_refined = True
    except ValueError:
        value = s_value
    except AttributeError:
        value = s_value
    return value, flag_refined


class ItemN(object):
    """Items data.

    It is internal class of cryspy library.
    You should use it only to create your own classes.
    """

    def __repr__(self):
        """
        Magic method print() is redefined.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        ls_out = [f"# Object '{self.get_name():}':\n"]
        res = str(self)
        if res == "":
            res = self.to_cif(flag_all_attributes=True)
        ls_out.append(res)

        method = self.methods_html()
        if method != "":
            ls_out.append(f"\n# Methods:\n{method:}\n")                
        return "\n".join(ls_out)

    def _repr_html_(self):
        """Representation in HTML format."""
        ls_html = [f"<h3>Object '{self.get_name():}'</h3>"]
        ls_html.append(self.attributes_to_html())

        report = self.report_html()
        if report != "":
            ls_html.append(f"<h3>Description </h3> {report:}")

        ls_html.append(f"<h3>Attributes and methods</h3>")
        try:
            names = sorted(self.ATTR_MANDATORY_NAMES)
            if len(names) != 0:
                ls_html.append("<b>Mandatory attributes: </b>")
                ls_html.append(f".{', .'.join(names):}<br>")
        except AttributeError:
            pass
        try:
            names = sorted(self.ATTR_OPTIONAL_NAMES)
            if len(names) != 0:
                ls_html.append("<b>Optional attributes: </b>")
                ls_html.append(f".{', .'.join(names):}<br>")
        except AttributeError:
            pass
        try:
            names = sorted(self.ATTR_INT_NAMES)
            if len(names) != 0:
                ls_html.append("<b>Internal attributes: </b>")
                ls_html.append(f".{', .'.join(names):}<br>")
        except AttributeError:
            pass
        try:
            names = sorted(self.ATTR_INT_PROTECTED_NAMES)
            if len(names) != 0:
                ls_html.append("<b>Internal protected attributes: </b>")
                ls_html.append(f".{', .'.join(names):}<br>")
        except AttributeError:
            pass

        method = self.methods_html()
        if method != "":
            ls_html.append(f"<b>Methods: </b> {method:}")

        return " ".join(ls_html)

    def methods_html(self):
        ls_html = [f".{func_name}" for func_name in
                   get_functions_of_objet(self)]
        return ", ".join(ls_html)+"."

    def __str__(self):
        """
        Magic method str() is redefined.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.to_cif()

    def __getattr__(self, name):
        """
        Magic method __getattr__ is slightly changed for special attributes.

        Parameters
        ----------
        name : TYPE
            Name of the attribute.

        Raises
        ------
        AttributeError
            Standard attribute error when it's necessary.

        Returns
        -------
        None.

        """
        if name.endswith("_as_string"):
            name_sh = name[:-(len("_as_string"))]
            try:
                val = getattr(self, name_sh)
            except AttributeError:
                return "."
            if val is None:
                return "."
            flag_ref = False
            keys = self.__dict__.keys()
            mark = ""
            if ((f"{name_sh:}_refinement" in keys) &
                    (f"{name_sh:}_sigma" in keys)):
                flag_ref = self.__dict__[f"{name_sh:}_refinement"]
                if (f"{name_sh:}_mark" in keys):
                    mark = self.__dict__[f"{name_sh:}_mark"]
            
            if flag_ref:
                val_sig = self.__dict__[f"{name_sh:}_sigma"]
                res = value_error_mark_to_string(val, val_sig, mark)
                
            else:
                s_format = None
                if (("D_FORMATS" in self.__dict__.keys()) |
                        ("D_FORMATS" in type(self).__dict__.keys())):
                    if name_sh in self.D_FORMATS.keys():
                        s_format = self.D_FORMATS[name_sh]
                if s_format is None:
                    res = f"{val:}"
                else:
                    res = s_format.format(val)
            return res
        if name in (self.ATTR_NAMES + self.ATTR_SIGMA
                    + self.ATTR_INT_PROTECTED_NAMES):
            raise AttributeError(f"Attribute '{name:}' is not defined in \
'{type(self).__name__:}'")
            # return None
        elif name in self.ATTR_INT_NAMES:
            if self.is_defined():
                # print("is_defined: ", self.is_defined())
                self.form_object()
                if name in  self.__dict__.keys():
                    return self.__dict__[name]
                else:
                    raise AttributeError(f"Attribute '{name:}' is not defined in \
'{type(self).__name__:}'")
            else:
                raise AttributeError(f"Attribute '{name:}' is not defined in \
'{type(self).__name__:}'")
                # return None
        else:
            raise AttributeError(f"Attribute '{name:}' is not defined in \
'{type(self).__name__:}'")

    def __setattr__(self, name: str, value) -> NoReturn:
        """
        Rules to set attribute.

        Parameters
        ----------
        name : str
            Name of attribute.
        value : TYPE
            Value of attribute.

        Returns
        -------
        None.

        """
        flag, flag_write = False, False
        flag_none_string = False
        if isinstance(value, str):
            flag_none_string = ((value == ".") | (value == "?"))

        if (value is None):
            flag, flag_write = True, True
            val_new = None
        elif flag_none_string:
            flag, flag_write = True, True
            val_new = None
        elif name in self.ATTR_NAMES:
            ind = self.ATTR_NAMES.index(name)
            val_type = self.ATTR_TYPES[ind]
            if ((val_type is bool) and (isinstance(value, str))):
                val_new = value == "True"
            else:
                val_new = val_type(value)
            flag, flag_write = True, True

            if name in self.D_CONSTRAINTS.keys():
                if val_new not in self.D_CONSTRAINTS[name]:
                    flag_write = False
            if name in self.D_MIN.keys():
                if val_new < self.D_MIN[name]:
                    val_new = self.D_MIN[name]
            if name in self.D_MAX.keys():
                if val_new > self.D_MAX[name]:
                    val_new = self.D_MAX[name]
            self.delete_internal_parameters()
        elif name in self.ATTR_SIGMA:
            val_new = float(value)
            flag, flag_write = True, True
        elif name in (self.ATTR_CONSTR_FLAG+self.ATTR_REF_FLAG):
            val_new = bool(value)
            flag, flag_write = True, True
        elif name in self.ATTR_CONSTR_MARK:
            val_new = str(value).strip()
            flag, flag_write = True, True

        if flag and flag_write:
            self.__dict__[name] = val_new
        elif not(flag):
            self.__dict__[name] = value

    def delete_internal_parameters(self):
        """
        Delete all internal parameters except protected one.

        Returns
        -------
        None.

        """
        l_del_name = []
        for key in self.__dict__.keys():
            flag = True
            if ((key == "D_MIN") | (key == "D_MAX")):
                flag = False
            elif key in (self.ATTR_NAMES + self.ATTR_SIGMA +
                         self.ATTR_CONSTR_FLAG + self.ATTR_REF_FLAG + self.ATTR_CONSTR_MARK +
                         self.ATTR_INT_PROTECTED_NAMES):
                flag = False
            if flag:
                l_del_name.append(key)
        for key in l_del_name:
            del self.__dict__[key]

    def is_attribute(self, name: str):
        """Give True if attribute is defined.
        
        Attention if parameter defined as None it's considered as defined.
        """
        flag = True
        try:
            getattr(self, name)
            # val = getattr(self, name)
            # if val is None:  # temporary solution
            #     flag = False
        except AttributeError or KeyError:
            flag = False
        return flag

    def is_defined(self) -> bool:
        """
        If all mandatory attributes is defined.

        Returns
        -------
        bool
            DESCRIPTION.

        """
        flag = True
        for name in self.ATTR_MANDATORY_NAMES:
            flag = self.is_attribute(name)
            if not(flag):
                break
        if (len(self.ATTR_MANDATORY_NAMES) == 0):
            flag = False
            for name in self.ATTR_OPTIONAL_NAMES:
                # print(name, getattr(self, name))
                if name in self.__dict__.keys():
                    flag = self.is_attribute(name)
                    if flag:
                        break
        return flag

    def form_object(self) -> NoReturn:
        """
        Calculate internal parameters for the object.

        Returns
        -------
        NoReturn
        """
        if ((type(self) == ItemN) and (len(self.ATTR_REF)==0)):
            l_ATTR_REF, l_ATTR_SIGMA = [], []
            l_ATTR_CONSTR_FLAG, l_ATTR_REF_FLAG = [], []
            l_ATTR_CONSTR_MARK = []
            variable_names = []
            for name in self.ATTR_NAMES:
                s_value = getattr(self, name)
                value, flag_refined = string_to_fixed_or_refined(s_value)
                if isinstance(value, float):
                    value, error, mark = string_to_value_error_mark(s_value)
                    name_sigma = f"{name:}_sigma"
                    name_constraint = f"{name:}_constraint"
                    name_refinement = f"{name:}_refinement"
                    name_mark = f"{name:}_mark"

                    self.__dict__[name] = value
                    self.__dict__[name_sigma] = error
                    self.__dict__[name_constraint] = False
                    self.__dict__[name_refinement] = flag_refined
                    self.__dict__[name_mark] = mark

                    l_ATTR_REF.append(name)
                    l_ATTR_SIGMA.append(name_sigma)
                    l_ATTR_CONSTR_FLAG.append(name_constraint)
                    l_ATTR_REF_FLAG.append(name_refinement)
                    l_ATTR_CONSTR_MARK.append(name_mark)
            self.__dict__["ATTR_REF"] = tuple(l_ATTR_REF)
            self.__dict__["ATTR_SIGMA"] = tuple(l_ATTR_SIGMA)
            self.__dict__["ATTR_CONSTR_FLAG"] = tuple(l_ATTR_CONSTR_FLAG)
            self.__dict__["ATTR_REF_FLAG"] = tuple(l_ATTR_REF_FLAG)
            self.__dict__["ATTR_CONSTR_MARK"] = tuple(l_ATTR_CONSTR_MARK)

    def to_cif(self, separator: str = "_", flag_all_attributes:bool=False) -> str:
        """
        Give data in CIF/STAR format.

        Parameters
        ----------
        separator : str, optional
            DESCRIPTION. The default is "_".

        Returns
        -------
        str
            DESCRIPTION.

        """
        ls_out = []
        prefix = self.PREFIX
        for name, name_cif in zip(self.ATTR_NAMES, self.ATTR_CIF):
            if name_cif == "":
                name_cif = name
            flag_value = self.is_attribute(name)
            if (flag_value | flag_all_attributes):
                if flag_value:
                    s_val = str(getattr(self, name))
                else:
                    s_val = "."
                l_s_val = s_val.split("\n")
                if len(l_s_val) > 1:
                    ls_out.append(f"_{prefix:}{separator:}{name_cif:}")
                    ls_out.append(";")
                    ls_out.extend(l_s_val)
                    ls_out.append(";")
                else:
                    if len(s_val.split(" ")) > 1:
                        ls_out.append(
                            f"_{prefix:}{separator:}{name_cif:} \"{s_val:}\"")
                    else:
                        s_val = getattr(self, f"{name:}_as_string")
                        ls_out.append(
                            f"_{prefix:}{separator:}{name_cif:} {s_val:}")
        return "\n".join(ls_out)

    def attributes_to_html(self) -> str:
        """Representation of defined parameters in HTML format.
        """
        ls_html = ["<table>"]
        ls_html.append("<tr><th>Attributes</th><th> Value </th></tr>")
        for name in self.ATTR_NAMES:
            flag_value = self.is_attribute(name)
            if flag_value:
                if name in self.ATTR_REF:
                    s_val = str(getattr(self, f"{name:}_as_string"))
                else:
                    s_val = str(getattr(self, name))
                if len(s_val.split("\n")) > 1:
                    ls_html.append(f"<tr><td>.{name:}</td><td> \
<i>multiline expression</i> </td></tr>")
                elif len(s_val)>40:                    
                    ls_html.append(f"<tr><td>.{name:}</td><td>{s_val[:40]:}...\
</td></tr>")
                else:
                    ls_html.append(f"<tr><td>.{name:}</td><td>{s_val:}\
</td></tr>")
        ls_html.append("</table>")
        return " ".join(ls_html)

    def get_name(self) -> str:
        """Get name."""
        return self.PREFIX.lower()

    def get_variable_names(self) -> list:
        """
        Get names of variable as a list.

        (((#prefix, #NAME), (#attribute, #index))

        Returns
        -------
        list
            List of names of variable.

        """
        if type(self) == ItemN:
            self.form_object()
        atr_ref = self.ATTR_REF
        atr_ref_flag = self.ATTR_REF_FLAG
        prefix = self.PREFIX
        variable_names = [((prefix, None), (name, None)) 
            for name, name_flag in
                zip(atr_ref, atr_ref_flag) if getattr(self, name_flag)]
        return variable_names

    def is_variables(self) -> bool:
        """Define is there variables or not."""
        atr_ref = self.ATTR_REF
        atr_ref_flag = self.ATTR_REF_FLAG
        flag = False
        for name, name_flag in zip(atr_ref, atr_ref_flag):
            if getattr(self, name_flag):
                flag = True
                break
        return flag

    def get_variable_by_name(self, name: tuple) -> Union[float, int, str]:
        """
        Get variable given by name.

        Parameters
        ----------
        name : tuple
            (((#prefix, ), (#attribute, ))

        Returns
        -------
        Union[float, int, str]
            Value.

        """
        
        prefix_t = name[0]
        if prefix_t[0] != self.PREFIX:
            return None
        if len(name) == 1:
            return self
        attr_t = name[1]
        attr_name = attr_t[0]
        return getattr(self, attr_name)

    def set_variable_by_name(self, name: tuple, value) -> NoReturn:
        """
        Set value to variable given by name.

        Parameters
        ----------
        name : tuple
            (((#prefix, ), (#attribute, ))
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        NoReturn

        """
        prefix_t, attr_t = name
        if prefix_t[0] != self.PREFIX:
            return
        attr_name = attr_t[0]
        setattr(self, attr_name, value)

    def _base_attributes_itemn(self, prefix, name_opt_cif, name_opt=None):
        name_mand, name_mand_cif = (), ()
        attr_ref = ()
        if name_opt is None:
            name_opt = name_opt_cif
        self.__dict__["PREFIX"] = prefix
        self.__dict__["ATTR_MANDATORY_CIF"] = name_mand_cif
        self.__dict__["ATTR_OPTIONAL_CIF"] = name_opt_cif
        self.__dict__["ATTR_REF"] = attr_ref
        self.__dict__["ATTR_CIF"] = name_mand_cif+name_opt_cif
        self.__dict__["ATTR_MANDATORY_NAMES"] = name_mand
        self.__dict__["ATTR_OPTIONAL_NAMES"] = name_opt
        self.__dict__["ATTR_NAMES"] = name_mand+name_opt
        types_mand = tuple(len(name_mand)*[str])
        types_opt = tuple(len(name_opt)*[str])
        self.__dict__["ATTR_MANDATORY_TYPES"] = types_mand
        self.__dict__["ATTR_OPTIONAL_TYPES"] = types_opt
        self.__dict__["ATTR_TYPES"] = types_mand + types_opt
        self.__dict__["ATTR_REF"] = ()
        self.__dict__["ATTR_SIGMA"] = ()
        self.__dict__["ATTR_CONSTR_FLAG"] = ()
        self.__dict__["ATTR_REF_FLAG"] = ()
        self.__dict__["ATTR_CONSTR_MARK"] = ()
        self.__dict__["D_FORMATS"] = {}
        self.__dict__["D_CONSTRAINTS"] = {}
        self.__dict__["D_DEFAULT"] = {}
        self.__dict__["ATTR_INT_NAMES"] = ()
        self.__dict__["D_MIN"] = {}
        self.__dict__["D_MAX"] = {}
        self.__dict__["ATTR_INT_PROTECTED_NAMES"] = (
            "PREFIX", "ATTR_MANDATORY_CIF", "ATTR_OPTIONAL_CIF",
            "ATTR_MANDATORY_NAMES", "ATTR_OPTIONAL_NAMES",
            "ATTR_REF", "ATTR_NAMES", "ATTR_CIF",
            "ATTR_MANDATORY_TYPES", "ATTR_TYPES", "ATTR_REF",
            "ATTR_SIGMA", "ATTR_CONSTR_FLAG", "ATTR_REF_FLAG", "ATTR_CONSTR_MARK",
            "D_FORMATS", "D_CONSTRAINTS", "D_DEFAULT",
            "ATTR_INT_NAMES", "ATTR_INT_NAMES",
            "ATTR_INT_PROTECTED_NAMES", "D_MIN", "D_MAX")

    @classmethod
    def from_cif(cls, string: str):
        """
        Create object of Cell class from string.

        Parameters
        ----------
        cls : TYPE
            DESCRIPTION.
        string : str
            string with descritption of the file.

        Returns
        -------
        ItemN
            DESCRIPTION.

        """
        cif_data = Data()
        flag = cif_data.take_from_string(string)
        item = None
        if cls is ItemN:
            l_name = [item.name for item in cif_data.items]
            if len(l_name) == 1:
                name = l_name[0]
                if name.find(".") != -1:
                    prefix = name.split(".")[0]
                else:
                    prefix = "_"+name.split("_")[1]
            else:
                prefix = find_prefix(*tuple(l_name))
            l_attr = [name[len(prefix)+1:] for name in l_name]

            prefix = prefix.strip("_").strip(".")
            name_opt_cif = tuple(l_attr)
            item = cls()
            item._base_attributes_itemn(prefix, name_opt_cif)
            prefix = item.PREFIX
            name_mand_cif = item.ATTR_MANDATORY_CIF
            name_opt_cif = item.ATTR_OPTIONAL_CIF
            name_mand = item.ATTR_MANDATORY_NAMES
            name_opt = item.ATTR_OPTIONAL_NAMES
            attr_ref = item.ATTR_REF
        else:
            prefix = cls.PREFIX
            name_mand_cif = cls.ATTR_MANDATORY_CIF
            name_opt_cif = cls.ATTR_OPTIONAL_CIF
            name_mand = cls.ATTR_MANDATORY_NAMES
            name_opt = cls.ATTR_OPTIONAL_NAMES
            attr_ref = cls.ATTR_REF

        name = name_mand + name_opt
        name_cif = name_mand_cif + name_opt_cif

        separator = "_"
        flag = all([cif_data.is_value("_" + prefix + separator + _attr) for
                    _attr in name_mand_cif])
        if not (flag):
            separator = "."
            flag = all(
                [cif_data.is_value("_" + prefix + separator + _attr) for _attr
                 in name_mand_cif])
        if flag:
            separator = "_"
            flag_2 = any([cif_data.is_value("_" + prefix + separator + _attr)
                          for _attr in name_cif])
            if not (flag_2):
                separator = "."
                flag_2 = any([cif_data.is_value("_" + prefix + separator +
                                                _attr) for _attr in name_cif])
            if flag_2:
                if item is None:
                    item = cls()
                for _attr, _cif_attr in zip(name, name_cif):
                    _cif_attr_full = "_" + prefix + separator + _cif_attr
                    if cif_data.is_value(_cif_attr_full):
                        if _attr in attr_ref:
                            value, error, mark = \
                                string_to_value_error_mark(
                                    cif_data[_cif_attr_full].value)
                            setattr(item, _attr, value)
                            if error is not None:
                                setattr(item, f"{_attr:}_sigma", error)
                                setattr(item, f"{_attr:}_refinement", True)
                                setattr(item, f"{_attr:}_mark", mark)
                            else:
                                setattr(item, f"{_attr:}_refinement", False)
                        else:
                            setattr(item, _attr,
                                    cif_data[_cif_attr_full].value)
            else:
                item = None
        else:
            item = None
        if item is not None:
            if item.is_defined():
                item.form_object()
        return item

    @classmethod
    def from_cif_file(cls, f_name: str):
        """Read from cif file."""
        if not(os.path.isfile(f_name)):
            raise UserWarning(f"File {f_name:} is not found.")
            return None
        str_from_cif = str(to_data(f_name))
        obj = cls.from_cif(str_from_cif)
        obj.file_input = f_name
        return obj

    def copy_from(self, obj):
        """Copy attributes from obj to self."""
        if type(obj) is not type(self):
            return
        for attr in (self.ATTR_NAMES + self.ATTR_SIGMA +
                     self.ATTR_CONSTR_FLAG + self.ATTR_REF_FLAG + self.ATTR_CONSTR_MARK):
            if obj.is_attribute(attr):
                setattr(self, attr, getattr(obj, attr))

    def report(self):
        return ""

    def report_html(self):
        return md_to_html(self.report())

    def plots(self):
        return []
    
    def fix_variables(self):
        """Fix variables."""
        for name in self.get_variable_names():
            name_ref = name[:-1] + ((f"{name[-1][0]:}_refinement",
                                     name[-1][1]), )
            self.set_variable_by_name(name_ref, False)

    def refine_all_variables(self):
        """Fix variables."""
        for name in self.ATTR_REF:
            name_constr = f"{name:}_constraint"
            constr = getattr(self, name_constr)
            if not(constr):
                name_ref = f"{name:}_refinement"
                setattr(self, name_ref, True)

    def set_variable(self, name: str, index=None):
        """Set refinement for variable given by name.
        
        Index parameters is used only for objects given as a matrix.
        """
        name_sh = name.strip(".").lower()
        if name_sh in self.ATTR_REF:
            flag_1 = self.is_attribute(name_sh)
            flag_2 = not(getattr(self, f"{name_sh:}_constraint"))
            if flag_1 & flag_2:
                setattr(self, f"{name_sh:}_refinement", True)
    
    def get_dictionary(self):
        res = {}
        return res
    
    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list=None, l_sigma: list=None):
        """
        """
        pass
        return None


# %%
import pytest
import os
import cryspy

L_RIETVELD_REFINEMENT_FILE = [
    os.path.join("rhochi_polarized_powder_1d_Ho2Ti2O7_5K_1T", "rhochi_polarized_powder_1d_Ho2Ti2O7_5K_1T.rcif"),
    os.path.join("rhochi_polarized_powder_2d_DyAl_5K_5T", "rhochi_polarized_powder_2d_DyAl_5K_5T.rcif"),
    os.path.join("rhochi_single_susceptibility_Ho2Ti2O7", "rhochi_single_susceptibility_Ho2Ti2O7.rcif"),
    os.path.join("rhochi_unpolarized_powder_1d_mcif_URu0.96Rh0.04Si2", "rhochi_unpolarized_powder_1d_mcif_URu0.96Rh0.04Si2.rcif"),
    os.path.join("rhochi_unpolarized_powder_1d_PbSO4", "rhochi_unpolarized_powder_1d_PbSO4.rcif"),
    os.path.join("rhochi_unpolarized_powder_1d_xrays_PbSO4", "rhochi_unpolarized_powder_1d_xrays_PbSO4.rcif"),
    os.path.join("rhochi_unpolarized_tof_powder_CeCuAl", "rhochi_unpolarized_tof_powder_CeCuAl.rcif"),
    os.path.join("only_for_test", "CuDSON.rcif"),
    os.path.join("only_for_test", "emso.rcif"),
    os.path.join("only_for_test", "Mg2Sb3Yb3O14.rcif"),
    os.path.join("only_for_test", "Mg2Sb3Ho3O14.rcif"),
    os.path.join("only_for_test", "mnbi4te7_single_lambdaover2.rcif"),
]

L_RIETVELD_REFINEMENT_CHI_SQ = [
    3273.82,
    67620.70,
    1135.42,
    2171.20,
    21016.14,
    6493.18,
    10716.91,
    104444.83,
     721829, 
     101.68,
     571.93,
     410.51,
]


   
def test_atom_multiplicity():
    f_dir_examples = os.path.join(os.path.dirname(os.path.dirname(__file__)), "examples")
    f_name = os.path.join(
        f_dir_examples,
        os.path.join("only_for_test",
                     "TbCo2Ni3.rcif"))
    rcif_obj = cryspy.load_file(f_name)
    cryspy.rhochi_no_refinement(rcif_obj)
    l_mult = rcif_obj.crystal_phase1.atom_site.multiplicity
    l_res = [1, 3, 3, 2, 2]
    for mult, res in zip(l_mult, l_res):
        assert mult == res
    


def test_rietveld_refinement():
    f_dir_examples = os.path.join(os.path.dirname(os.path.dirname(__file__)), "examples")
    # %%
    for f_file, chi_sq in zip(L_RIETVELD_REFINEMENT_FILE, L_RIETVELD_REFINEMENT_CHI_SQ):
        f_name = os.path.join(f_dir_examples, f_file)

        rcif_obj = cryspy.load_file(f_name)

        res = cryspy.rhochi_rietveld_refinement(rcif_obj)
        
        assert res["chi_sq"] < chi_sq



def test_mem():
    f_dir_examples = os.path.join(os.path.dirname(os.path.dirname(__file__)), "examples")
    f_name = os.path.join(
        f_dir_examples,
        os.path.join("mempy_spin_density_YTiO3",
                     "mempy_spin_density_YTiO3.rcif"))
    try:
        rcif_obj = cryspy.load_file(f_name)
        cryspy.mempy_spin_density_reconstruction(rcif_obj)
        assert True
    except:
        assert False

    f_name = os.path.join(
        f_dir_examples,
        os.path.join("mempy_magnetization_density_Yb2Ti2O7_2K_1T",
                     "mempy_magnetization_density_Yb2Ti2O7_2K_1T.rcif"))
    try:
        rcif_obj = cryspy.load_file(f_name)
        cryspy.mempy_magnetization_density_reconstruction(rcif_obj)
        assert True
    except:
        assert False



from typing import NoReturn
import numpy
import matplotlib.pyplot as plt

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class TOFMeas(ItemN):
    """Measured data for time-of-flight experiments.

    Attributes
    ----------
        - time (mandatory)
        - intensity_plus, intensity_plus_sigma, intensity_minus,
          intensity_minus_sigma, intensity, intensity_sigma (optional)

    time is given in microseconds.
    """
    ATTR_MANDATORY_NAMES = ("time", )
    ATTR_MANDATORY_TYPES = (float, )
    ATTR_MANDATORY_CIF = ("time", )

    ATTR_OPTIONAL_NAMES = ("intensity_plus", "intensity_plus_sigma",
                           "intensity_minus", "intensity_minus_sigma",
                           "intensity", "intensity_sigma")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float)
    ATTR_OPTIONAL_CIF = ("intensity_plus", "intensity_plus_sigma",
                         "intensity_minus", "intensity_minus_sigma",
                         "intensity", "intensity_sigma")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {'intensity_plus': "{:.2f}", 'intensity_plus_sigma': "{:.2f}",
                 'intensity_minus': "{:.2f}", 'intensity_minus_sigma': "{:.2f}",
                 'intensity': "{:.2f}", 'intensity_sigma': "{:.2f}",
                 'time': "{:.3f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_meas"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFMeas, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"time": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def apply_constraints(self):
        """Apply constraints."""
        keys = self.__dict__.keys()
        if (("intensity_plus" in keys) & ("intensity_plus_sigma" not in keys)):
            self.intensity_plus_sigma = (self.intensity_plus)**0.5
        if (("intensity_minus" in keys) & ("intensity_minus_sigma" not in keys)):
            self.intensity_minus_sigma = (self.intensity_minus)**0.5
        if (("intensity" in keys) & ("intensity_sigma" not in keys)):
            self.intensity_sigma = (self.intensity)**0.5


class TOFMeasL(LoopN):
    """Measured data for time-of-flight experiments.

    This section contains the measured diffractogram and information
    about the conditions used for the measurement of the diffraction 
    data set, prior to processing and application of correction
    terms. While additional information may be added to the CIF
    as data are processed and transported between laboratories
    (possibly with the addition of a new _pd_block_id entry), the
    information in this section of the CIF will rarely be changed
    once data collection is complete.

    Where possible, measurements in this section should have no
    post-collection processing applied (normalizations, corrections,
    smoothing, zero-offset corrections etc.). Such corrected
    measurements should be recorded in the _pd_proc_ section.

    Data sets that are measured as counts, where a standard
    uncertainty can be considered equivalent to the standard
    deviation and where the standard deviation can be estimated
    as the square root of the number of counts recorded, should
    use the _pd_meas_counts_ fields. All other intensity values
    should be recorded using _pd_meas_intensity_.
    """

    ITEM_CLASS = TOFMeas
    ATTR_INDEX = "time"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFMeasL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def is_polarized(self):
        """Give True if polaraized data are defined."""
        flag = False
        if len(self.items) != 0:
            items_0 = self.items[0]
            flag_up = items_0.is_attribute("intensity_plus")
            flag_down = items_0.is_attribute("intensity_minus")
            flag = flag_up & flag_down
        return flag

    def apply_constraints(self):
        """Redefined applyied constraints."""
        for item in self.items:
            item.apply_constraints()

    def plots(self):
        return [self.plot_up_down(), self.plot_sum(), self.plot_diff()]

    def plot_up_down(self):
        """Plot experimental intensity up and down vs. 2 theta (degrees)
        """
        fig, ax = plt.subplots()
        ax.set_title("I_up and I_down")
        ax.set_xlabel("Time (microseconds)")
        ax.set_ylabel('Intensity')

        if (self.is_attribute("time") & self.is_attribute("intensity_plus") & 
            self.is_attribute("intensity_plus_sigma") &
            self.is_attribute("intensity_minus") & 
            self.is_attribute("intensity_minus_sigma")):
            np_time = numpy.array(self.time, dtype=float)
            np_up = numpy.array(self.intensity_plus, dtype=float)
            np_sup = numpy.array(self.intensity_plus_sigma, dtype=float)
            np_down = numpy.array(self.intensity_minus, dtype=float)
            np_sdown = numpy.array(self.intensity_minus_sigma, dtype=float)

            ax.fill_between(np_time, (np_up-np_sup), (np_up+np_sup), color="r", alpha=0.4, label="plus")
            ax.fill_between(np_time, (np_down-np_sdown), (np_down+np_sdown), color="b", alpha=0.4, label="minus")

            # ax.plot(np_time, np_up, "r-", alpha=0.2)
            # ax.errorbar(np_time, np_up, yerr=np_sup, fmt="ro", alpha=0.2,
            #             label="up")
            # ax.plot(np_time, np_down, "b-", alpha=0.2)
            # ax.errorbar(np_time, np_down, yerr=np_sdown, fmt="bs", alpha=0.2,
            #             label="down")
        else:
            return
        ax.legend(loc='upper right')
        fig.tight_layout()
        return (fig, ax)
    
    def plot_sum(self):
        """Plot experimental unpolarized intensity vs. time
        """
        fig, ax = plt.subplots()
        ax.set_title("Unpolarized intensity: I_up + I_down")
        ax.set_xlabel("Time (microseconds)")
        ax.set_ylabel('Intensity')

        if (self.is_attribute("time") & self.is_attribute("intensity_plus") & 
            self.is_attribute("intensity_plus_sigma") &
            self.is_attribute("intensity_minus") & 
            self.is_attribute("intensity_minus_sigma")):
            np_time = numpy.array(self.time, dtype=float)
            np_up = numpy.array(self.intensity_plus, dtype=float)
            np_sup = numpy.array(self.intensity_plus_sigma, dtype=float)
            np_down = numpy.array(self.intensity_minus, dtype=float)
            np_sdown = numpy.array(self.intensity_minus_sigma, dtype=float)
            np_sum = np_up + np_down
            np_ssum = numpy.sqrt(numpy.square(np_sup)+numpy.square(np_sdown))
            ax.fill_between(np_time, (np_sum-np_ssum), (np_sum+np_ssum), color="k", alpha=0.4, label="experiment")
            # ax.plot(np_time, np_sum, "k-", alpha=0.2)
            # ax.errorbar(np_time, np_sum, yerr=np_ssum, fmt="ko", alpha=0.2,
            #             label="experiment")
        elif (self.is_attribute("time") & self.is_attribute("intensity") & 
              self.is_attribute("intensity_sigma")):
            np_time = numpy.array(self.time, dtype=float)
            np_sum = numpy.array(self.intensity, dtype=float)
            np_ssum = numpy.array(self.intensity_sigma, dtype=float)
            ax.fill_between(np_time, (np_sum-np_ssum), (np_sum+np_ssum), color="k", alpha=0.4, label="experiment")
            # ax.plot(np_time, np_sum, "k-", alpha=0.2)
            # ax.errorbar(np_time, np_sum, yerr=np_ssum, fmt="ko", alpha=0.2,
            #             label="experiment")
        ax.legend(loc='upper right')
        fig.tight_layout()
        return (fig, ax)

    def plot_diff(self):
        """Plot experimental polarized intensity vs. time
        """
        if not(self.is_attribute("time") & self.is_attribute("intensity_plus") & 
               self.is_attribute("intensity_plus_sigma") &
               self.is_attribute("intensity_minus") & 
               self.is_attribute("intensity_minus_sigma")):
            return
        fig, ax = plt.subplots()
        ax.set_title("Polarized intensity: I_up - I_down")
        ax.set_xlabel("Time (microseconds)")
        ax.set_ylabel('Intensity')
            
        np_time = numpy.array(self.time, dtype=float)
        np_up = numpy.array(self.intensity_plus, dtype=float)
        np_sup = numpy.array(self.intensity_plus_sigma, dtype=float)
        np_down = numpy.array(self.intensity_minus, dtype=float)
        np_sdown = numpy.array(self.intensity_minus_sigma, dtype=float)
        np_diff = np_up - np_down
        np_sdiff = numpy.sqrt(numpy.square(np_sup)+numpy.square(np_sdown))

        ax.plot([np_time.min(), np_time.max()], [0., 0.], "k:")
        ax.fill_between(np_time, (np_diff-np_sdiff), (np_diff+np_sdiff), color="k", alpha=0.4, label="experiment")

        # ax.plot(np_time, np_diff, "k-", alpha=0.2)
        # ax.errorbar(np_time, np_diff, yerr=np_sdiff, fmt="ko", alpha=0.2,
        #             label="experiment")
        ax.legend(loc='upper right')
        fig.tight_layout()
        return (fig, ax)


# s_cont = """
#   loop_
#   _tof_meas_time
#   _tof_meas_intensity_plus
#   _tof_meas_intensity_plus_sigma
#   _tof_meas_intensity_minus
#   _tof_meas_intensity_minus_sigma
# 3001.589    40409.0    462.0    40409.0    462.0    
# 3003.09     40171.0    460.0    40409.0    462.0    
# 3004.591    39733.0    458.0   40409.0    462.0    
# """

# obj = TOFMeasL.from_cif(s_cont)
# print(obj, end="\n\n")



from typing import NoReturn

import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary
from cryspy.A_functions_base.function_1_strings import \
    transform_string_to_r_b

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_space_group_symop_magn_centering import \
    SpaceGroupSymopMagnCentering


class SpaceGroupSymopMagnOperation(ItemN):
    """Magnetic space-group symmetry operation.

    Attributes
    ----------
    - xyz (mandatory)
    - description, id (optioanl)
    """

    ATTR_MANDATORY_NAMES = ("xyz", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("xyz", )

    ATTR_OPTIONAL_NAMES = ("id", "description")
    ATTR_OPTIONAL_TYPES = (str, str)
    ATTR_OPTIONAL_CIF = ("id", "description")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ("r", "b", "r_11", "r_12", "r_13", "r_21", "r_22",
                      "r_23", "r_31", "r_32", "r_33", "b_1", "b_2", "b_3",
                      "theta", "sym_elem")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "space_group_symop_magn_operation"

    def __init__(self, **kwargs) -> NoReturn:
        super(SpaceGroupSymopMagnOperation, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        """Form object."""
        xyz = self.xyz
        if xyz is None:
            return False
        r, b = transform_string_to_r_b(xyz, labels=("x", "y", "z"))
        r_11, r_12, r_13 = int(r[0, 0]), int(r[0, 1]), int(r[0, 2])
        r_21, r_22, r_23 = int(r[1, 0]), int(r[1, 1]), int(r[1, 2])
        r_31, r_32, r_33 = int(r[2, 0]), int(r[2, 1]), int(r[2, 2])
        den_1, den_2 = b[0].denominator, b[1].denominator
        den_3 = b[2].denominator
        num_1, num_2 = b[0].numerator, b[1].numerator
        num_3 = b[2].numerator
        den = numpy.lcm.reduce([den_1, den_2, den_3])
        num_1 *= den//den_1
        num_2 *= den//den_2
        num_3 *= den//den_3

        theta = int(b[3])
        self.__dict__["r_11"] = r_11
        self.__dict__["r_12"] = r_12
        self.__dict__["r_13"] = r_13
        self.__dict__["r_21"] = r_21
        self.__dict__["r_22"] = r_22
        self.__dict__["r_23"] = r_23
        self.__dict__["r_31"] = r_31
        self.__dict__["r_32"] = r_32
        self.__dict__["r_33"] = r_33
        self.__dict__["b_1"] = b[0]
        self.__dict__["b_2"] = b[1]
        self.__dict__["b_3"] = b[2]
        self.__dict__["theta"] = theta
        self.__dict__["sym_elem"] = numpy.array([
            num_1, num_2, num_3, den,
            r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33,
            theta], dtype=int)
        # self.__dict__["sym_elem"] = numpy.array([
        #     num_1, num_2, num_3, den,
        #     r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33,
        #     theta*r_11, theta*r_12, theta*r_13, theta*r_21, theta*r_22,
        #     theta*r_23, theta*r_31, theta*r_32, theta*r_33], dtype=int)

    def get_symop_magn_operation_by_magn_centering(
            self, space_group_symop_magn_centering:
            SpaceGroupSymopMagnCentering):
        """Get symop magn operations by symop magn centering."""
        pass


class SpaceGroupSymopMagnOperationL(LoopN):
    """Magnetic space-group symmetry operations."""

    ITEM_CLASS = SpaceGroupSymopMagnOperation
    ATTR_INDEX = "id"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(SpaceGroupSymopMagnOperationL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def get_sym_elems(self):
        """Get sym elems."""
        res = numpy.array([item.sym_elem for item in self.items], dtype=int
                          ).transpose()
        return res

# s_cont = """
# loop_
# _space_group_symop_magn_operation_id
# _space_group_symop_magn_operation_xyz
# 1    x,y,z,+1
# 2    -y,x-y+1/2,z,+1
# 3    -x+y+1/2,-x,z,+1
# 4    x-y+1/2,-y,-z,+1
# 5    y,x,-z,+1
# 6    -x,-x+y+1/2,-z,+1
# 7    -x+y+1/2,-x,-z,-1
# 8    x,y,-z,-1
# 9    -y,x-y+1/2,-z,-1
# 10   -x,-x+y+1/2,z,-1
# 11   x-y+1/2,-y,z,-1
# 12   y,x,z,-1
# """

# obj = SpaceGroupSymopMagnOperationL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["3"], end="\n\n")
# print(obj.get_sym_elems(), end="\n\n")
# print(obj.get_sym_elems().shape)




"""AtomSiteMoment, AtomSiteMomentL classes are given."""
import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class AtomSiteMoment(ItemN):
    """AtomSiteMoment class.

    This category provides a loop for presenting the magnetic moments
    of atoms in one of several coordinate systems. This is a child category
    of the AtomSite category, so that the magnetic moments can either be
    listed alongside the non-magnetic atom properties in the main AtomSite loop
    (not realized) or be listed in a separate loop (realized)

    Attributes
    ----------
        - label, symmform (mandatory)
        - cartn_x, cartn_y, cartn_z, crystalaxis_x, crystalaxis_y,
          crystalaxis_z, modulation_flag, refinement_flags_magnetic,
          spherical_azimuthal, spherical_modulus, spherical_polar
    """

    ATTR_MANDATORY_NAMES = ("label",)
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("label", )

    ATTR_OPTIONAL_NAMES = (
        "cartn_x", "cartn_y", "cartn_z", "crystalaxis_x", "crystalaxis_y",
        "crystalaxis_z",  "symmform", "modulation_flag", "refinement_flags_magnetic",
        "spherical_azimuthal", "spherical_modulus", "spherical_polar")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float, str,
                           str, str, float, float, float)
    ATTR_OPTIONAL_CIF = (
        "cartn_x", "cartn_y", "cartn_z", "crystalaxis_x", "crystalaxis_y",
        "crystalaxis_z",  "symmform", "modulation_flag", "refinement_flags_magnetic",
        "spherical_azimuthal", "spherical_modulus", "spherical_polar")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("crystalaxis_x", "crystalaxis_y", "crystalaxis_z")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {"modulation_flag": ["yes", "y", "no", "n"],
                     "refinement_flags_magnetic": ["S", "M", "A", "SM", "SA",
                                                   "MA", "SMA"]}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_site_moment"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomSiteMoment, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def moment(self):
        """Calculate moment.

        FIXME: It is valid only for cubic crystals
        """
        np_x = numpy.array(self.crystalaxis_x, dtype=float)
        np_y = numpy.array(self.crystalaxis_y, dtype=float)
        np_z = numpy.array(self.crystalaxis_z, dtype=float)
        np_moment = numpy.sqrt(numpy.square(np_x) + numpy.square(np_y) +
                               numpy.square(np_z))
        return np_moment

    def calc_zeeman(self, field_cryst):
        """Calc Zeeman energy.

        FIXME: It is valid only for cubic crystals
        moment*sin(moment^field) at angle decreesing
        """
        h_a, h_b, h_c = field_cryst[0], field_cryst[1], field_cryst[2]

        if abs(h_a)+abs(h_b)+abs(h_c) == 0.:
            np_val = numpy.zeros(len(self.label), dtype=float)
            return np_val
        else:
            mod_h = (abs(h_a)**2+abs(h_b)**2+abs(h_c)**2)**0.5
            h_a_n, h_b_n, h_c_n = h_a/mod_h, h_b/mod_h, h_c/mod_h
            np_moment_sq = numpy.square(self.moment)
            np_x = numpy.array(self.crystalaxis_x, dtype=float)
            np_y = numpy.array(self.crystalaxis_y, dtype=float)
            np_z = numpy.array(self.crystalaxis_z, dtype=float)
            np_val = h_a_n*np_x+h_b_n*np_y+h_c_n*np_z
            np_res = numpy.sqrt(np_moment_sq-numpy.square(np_val))
        return np_res

    def to_cif(self, separator: str = ".") -> str:
        return super(AtomSiteMoment, self).to_cif(separator=separator)

class AtomSiteMomentL(LoopN):
    """AtomSiteMomentL class.

    AtomSiteMomentL category provides a loop for presenting the magnetic
    moments  of atoms in one of several coordinate systems. This is a child
    category of the AtomSite category, so that the magnetic moments can either
    be listed alongside the non-magnetic atom properties in the main AtomSite
    loop (not realized) or be listed in a separate loop (realized)
    """

    ITEM_CLASS = AtomSiteMoment
    ATTR_INDEX = "label"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomSiteMomentL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def to_cif(self, separator: str = ".") -> str:
        return super(AtomSiteMomentL, self).to_cif(separator=separator)

# s_cont = """
# loop_
# _atom_site_moment_label
# _atom_site_moment_crystalaxis_x
# _atom_site_moment_crystalaxis_y
# _atom_site_moment_crystalaxis_z
# _atom_site_moment_symmform
# Tm1_1   -6.44   0.0     0.0   mx,0,0
# Tm1_2   -6.44   0.0     0.0   mx,0,0
# Tm1_3   -6.44   -6.44   0.0   mx,my,0
#   """

# obj = AtomSiteMomentL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["Tm1_2"], end="\n\n")



import numpy
import scipy
import scipy.special

from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


# FIXME: estimate d by time for epithermal neutrons
class TOFParameters(ItemN):
    """Parameters of the reflexion positions in time-of-flight experiments.

    Attributes
    ----------
        - zero, dtt1, ttheta_bank (mandatory)
        - neutrons, dtt2, zerot, dtt1t, dtt2t, width, x_cross, field,
          exinction (optional)

    for thermal neutrons
    
        time = zero + dtt1 * d + dtt2 * d**2

    or for epithermal neutrons
    
        time_e = zero + dtt1 * d 
        
        time_t = zerot + dtt1t * d - dtt2t / d 
        
        n_cross = 0.5*erfc(Width * (x_cross - 1/d))
        
        time = n_cross * time_e + (1-n_cross) time_t

    time is given in microseconds.
    """
    ATTR_MANDATORY_NAMES = ("zero", "dtt1", "ttheta_bank")
    ATTR_MANDATORY_TYPES = (float, float, float)
    ATTR_MANDATORY_CIF = ("Zero", "Dtt1", "2theta_bank")

    ATTR_OPTIONAL_NAMES = ('neutrons', "dtt2", "zerot", "dtt1t", "dtt2t",
                           "width", "x_cross", "field", "extinction")
    ATTR_OPTIONAL_TYPES = (str, float, float, float, float, float, float,
                           float, float)
    ATTR_OPTIONAL_CIF = ('neutrons', "dtt2", "zerot", "dtt1t", "dtt2t",
                         "width", "x_cross", "field", "extinction")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("zero", "dtt1", "dtt2", "zerot", "dtt1t", "dtt2t")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {
        "zero": "{:.5f}", "dtt1": "{:.5f}", "dtt2": "{:.5f}",
        "zerot": "{:.5f}", "dtt1t": "{:.5f}", "dtt2t": "{:.5f}",
        "ttheta_bank": "{:.2f}", "width": "{:.2f}", "x_cross": "{:.2f}",
        "field": "{:.3f}", "extinction": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {"neutrons": ["thermal", "epithermal"]}

    # default values for the parameters
    D_DEFAULT = {"zero": 0., "neutrons": "thermal"}

    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_parameters"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFParameters, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_time_by_d(self, d):
        """Calculate time by given d
        
        time = zero + dtt1 * d + dtt2 * d**2
        (if dtt2 is defined)

        time = zero + dtt1 * d
        (if dtt2 is not defined)

        Parameters
        ----------
        d : TYPE
            DESCRIPTION.

        Returns
        -------
        time : TYPE
            DESCRIPTION.

        """
        if self.neutrons == "epithermal":
            time_e = self.zero + self.dtt1 * d
            time_t = self.zerot + self.dtt1t * d - self.dtt2t / d 
            n_cross = 0.5*scipy.special.erfc(self.width * (self.x_cross - 1/d))
            time = n_cross * time_e + (1-n_cross) * time_t
        else:  # self.neutrons == "thermal"
            time = self.zero + self.dtt1 * d + self.dtt2 * d**2
        return time

    def calc_d_min_max(self, time):
        """Calculate d_min, d_max 
        

        Parameters
        ----------
        time : TYPE
            DESCRIPTION.

        Returns
        -------
        d_min : TYPE
            DESCRIPTION.
        d_max : TYPE
            DESCRIPTION.

        """
        time_min = numpy.min(time)
        time_max = numpy.max(time)
        if self.neutrons == "epithermal":
            raise AttributeError("function calc_d_min_max is not realized for \
epithermal neutrons")
            d_min = (time_min-self.zero)/self.dtt1
            d_max = (time_max-self.zero)/self.dtt1
        else:  # self.neutrons == "thermal"
            det_sq_min = self.dtt1**2 - 4.*self.dtt2*(self.zero - time_min)
            det_sq_max = self.dtt1**2 - 4.*self.dtt2*(self.zero - time_max)
            d_max = (-self.dtt1+det_sq_max**0.5)/(2.*self.dtt2)
            d_min = (-self.dtt1+det_sq_min**0.5)/(2.*self.dtt2)
        return d_min, d_max
    
    def calc_d_by_time(self, time):
        """Calculate d by given time 
        
        Relation between d and time is
        
        time = zero + dtt1 * d + dtt2 * d**2
        (if dtt2 is defined)

        time = zero + dtt1 * d
        (if dtt2 is not defined)

        Parameters
        ----------
        time : TYPE
            DESCRIPTION.

        Returns
        -------
        d : TYPE
            DESCRIPTION.

        """
        if self.neutrons == "epithermal":
            raise AttributeError("function calc_d_min_max is not realized for \
epithermal neutrons")
            d = (time - self.zero)/self.dtt1
        else:  # self.neutrons == "thermal"
            det = numpy.sqrt(self.dtt1**2 - 4.*(self.zero-time)*self.dtt2)
            if self.dtt2 < 0.:
                d = (det-self.dtt1)/(2.*self.dtt2)
            elif self.dtt2 > 0.:
                d = (-det-self.dtt1)/(2.*self.dtt2)
            else:
                d = (time - self.zero)/self.dtt1
        return d

    def calc_time_by_sthovl(self, sthovl):
        """Calculate time by given sin(theta)/lambda
        
        Returns
        -------
        time : TYPE
            DESCRIPTION.

        """
        d = 0.5/sthovl
        time = self.calc_time_by_d(d)
        return time

    def get_dictionary(self):
        res = {}
        res["ttheta_bank"] = getattr(self, "ttheta_bank") * numpy.pi/180
        res["neutron_type"] = getattr(self, "neutrons")
        res["zero"] = numpy.array([getattr(self, "zero"), ], dtype=float)
        res["flags_zero"] = numpy.array([getattr(self, "zero_refinement"), ], dtype=bool)
        res["dtt1"] = numpy.array([getattr(self, "dtt1"), ], dtype=float)
        res["flags_dtt1"] = numpy.array([getattr(self, "dtt1_refinement"), ], dtype=bool)
        if self.is_attribute("dtt2"):
            res["dtt2"] = numpy.array([getattr(self, "dtt2"), ], dtype=float)
            res["flags_dtt2"] = numpy.array([getattr(self, "dtt2_refinement"), ], dtype=bool)
        if self.is_attribute("zerot"):
            res["zerot"] = numpy.array([getattr(self, "zerot"), ], dtype=float)
            res["flags_zerot"] = numpy.array([getattr(self, "zerot_refinement"), ], dtype=bool)
        if self.is_attribute("dtt1t"):
            res["dtt1t"] = numpy.array([getattr(self, "dtt1t"), ], dtype=float)
            res["flags_dtt1t"] = numpy.array([getattr(self, "dtt1t_refinement"), ], dtype=bool)
        if self.is_attribute("dtt2t"):
            res["dtt2t"] = numpy.array([getattr(self, "dtt2t"), ], dtype=float)
            res["flags_dtt2t"] = numpy.array([getattr(self, "dtt2t_refinement"), ], dtype=bool)
        return res

class TOFParametersL(LoopN):
    """Parameters of the reflexion positions in time-of-flight experiments.

    """
    ITEM_CLASS = TOFParameters
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFParametersL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name
   

# s_cont = """
# _tof_parameters_zero 2.921
# _tof_parameters_dtt1 6167.247
# _tof_parameters_dtt2 -2.280
# _tof_parameters_2theta_bank 145.
# """

# obj = TOFParameters.from_cif(s_cont)
# print(obj, end="\n\n")



"""Description of Crystal class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"
import numpy

from typing import NoReturn

from cryspy.A_functions_base.database import DATABASE
from cryspy.A_functions_base.charge_form_factor import calc_jl_for_ion, calc_scattering_amplitude_tabulated, get_atom_name_ion_charge_shell, get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol
from cryspy.A_functions_base.debye_waller_factor import calc_param_iso_aniso_by_b_iso_beta, calc_u_ij_by_beta
from cryspy.A_functions_base.matrix_operations import calc_m1_m2_inv_m1, calc_m_v
from cryspy.A_functions_base.magnetic_form_factor import get_j0_j2_parameters
from cryspy.A_functions_base.unit_cell import calc_eq_ccs_by_unit_cell_parameters, calc_m_m_by_unit_cell_parameters, calc_reciprocal_by_unit_cell_parameters
from cryspy.A_functions_base.structure_factor import calc_f_nucl_by_dictionary, calc_sft_ccs_by_dictionary, calc_f_m_perp_by_sft, calc_bulk_susceptibility_by_dictionary
from cryspy.A_functions_base.symmetry_elements import calc_full_mag_elems, calc_symm_flags, define_bravais_type_by_symm_elems
from cryspy.A_functions_base.symmetry_constraints import calc_sc_beta, calc_sc_fract_sc_b, calc_sc_chi, calc_sc_chi_full

from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.preocedures import take_items_by_class

from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSite, AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_type import AtomTypeL
from cryspy.C_item_loop_classes.cl_1_atom_site_aniso import \
    AtomSiteAnisoL
from cryspy.C_item_loop_classes.cl_1_atom_site_susceptibility import \
    AtomSiteSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_atom_site_exchange import AtomSiteExchangeL
from cryspy.C_item_loop_classes.cl_1_atom_type_scat import \
    AtomTypeScatL
from cryspy.C_item_loop_classes.cl_2_atom_site_scat import \
    AtomSiteScatL
from cryspy.C_item_loop_classes.cl_1_atom_site_moment import AtomSiteMoment, AtomSiteMomentL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_atom_local_axes import \
    AtomLocalAxesL
from cryspy.C_item_loop_classes.cl_1_atom_electron_configuration \
    import AtomElectronConfigurationL
from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_space_group_symop_magn_centering import \
    SpaceGroupSymopMagnCenteringL

from cryspy.C_item_loop_classes.cl_2_atom_rho_orbital_radial_slater \
    import AtomRhoOrbitalRadialSlaterL
from cryspy.C_item_loop_classes.cl_2_space_group_symop_magn_operation import \
    SpaceGroupSymopMagnOperationL
from cryspy.C_item_loop_classes.cl_2_space_group import SpaceGroup





from cryspy.D_functions_item_loop.function_1_report_magnetization_ellipsoid \
    import magnetization_ellipsoid_by_u_ij, report_main_axes_of_magnetization_ellipsoids

na = numpy.newaxis

class Crystal(DataN):
    """
    Crystal structure description.

    Data items in the CRYSTAL category record details about
    crystal structure.

    Methods
    -------
        - calc_b_iso_beta
        - calc_f_nucl
        - calc_susceptibility_moment_tensor
        - calc_magnetic_moments_with_field_loc
        - report_main_axes_of_magnetization_ellipsoids
        - calc_main_axes_of_magnetization_ellipsoids
        - calc_magnetization_ellipsoid
        - calc_hkl_in_range
        - calc_hkl
        - calc_refln_susceptibility
        - calc_refln

    Attributes
    ----------
        - space_group, cell, atom_site (mandatory)
        - atom_type, atom_site_aniso, atom_site_susceptibility,
          atom_site_scat, atom_type_scat, atom_local_axes,
          atom_electron_confiduration (optional)
    """

    CLASSES_MANDATORY = (Cell, AtomSiteL,)
    CLASSES_OPTIONAL = (SpaceGroup,
        AtomTypeL, AtomSiteAnisoL, AtomSiteSusceptibilityL, AtomSiteExchangeL, AtomSiteScatL,
        AtomTypeScatL, AtomLocalAxesL, AtomElectronConfigurationL,
        AtomSiteMomentL, SpaceGroupSymopMagnOperationL, SpaceGroupSymopMagnCenteringL)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "crystal"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(Crystal, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        """Redefined method of DataN."""
        self.apply_constraints()

    def apply_constraints(self) -> NoReturn:
        """
        Symmetry constraints on parameters.

        Returns
        -------
        NoReturn
            DESCRIPTION.

        """
        if self.is_attribute("space_group"):
            space_group = self.space_group
            space_group_wyckoff = space_group.space_group_wyckoff
    
            cell = self.cell
            cell.type_cell = space_group.bravais_type
            cell.it_coordinate_system_code = space_group.it_coordinate_system_code
            cell.apply_constraints()
    
            atom_site = self.atom_site
            atom_site.apply_constraints(space_group_wyckoff)
    
            if self.is_attribute("atom_site_aniso"):
                atom_site_aniso = self.atom_site_aniso
                atom_site_aniso.apply_space_group_constraint(
                    atom_site, space_group)
    
            if self.is_attribute("atom_site_susceptibility"):
                atom_site_susceptibility = self.atom_site_susceptibility
                atom_site_susceptibility.apply_chi_iso_constraint(cell)
                atom_site_susceptibility.apply_space_group_constraint(
                    atom_site, space_group, cell)
            if self.is_attribute("atom_site_exchange"):
                atom_site_exchange = self.atom_site_exchange
                atom_site_exchange.apply_j_iso_constraint()
                atom_site_exchange.apply_space_group_constraint(
                    atom_site, space_group, cell)

    def calc_b_iso_beta(self):
        """
        Calculate b_iso and beta_ij based on atom_site and atom_sites.

        For each atom defined in atom_site.
        """
        a_s = self.atom_site
        try:
            a_s_a = self.atom_site_aniso
        except AttributeError:
            a_s_a = None
        l_b_iso, l_beta = [], []
        coeff = float(8.*numpy.pi**2)
        cell = self.cell
        for item_a_s in a_s.items:
            label_atom = item_a_s.label
            try:
                adp_type = item_a_s.adp_type
            except AttributeError:
                adp_type = None
            b_iso = 0.
            beta = (0., 0., 0., 0., 0., 0.)
            if adp_type == "Uiso":
                u_iso = float(item_a_s.u_iso_or_equiv)
                b_iso = float(8.*numpy.pi**2*u_iso)
            elif adp_type == "Biso":
                b_iso = float(item_a_s.b_iso_or_equiv)
            elif adp_type == "Uovl":
                # FIXME: correct it
                u_iso = float(item_a_s.u_iso_or_equiv)
                b_iso = coeff*u_iso
            elif adp_type == "Umpe":
                # FIXME: correct it
                u_iso = float(item_a_s.u_iso_or_equiv)
                b_iso = float(8.*numpy.pi**2*u_iso)
            elif adp_type == "Uani":
                item_a_s_a = a_s_a[label_atom]
                beta = item_a_s_a.calc_beta(cell)
            elif adp_type == "Bovl":
                # FIXME: correct it
                b_iso = float(item_a_s.b_iso_or_equiv)
            elif adp_type == "Bani":
                item_a_s_a = a_s_a[label_atom]
                beta = (float(item_a_s_a.b_11), float(item_a_s_a.b_22),
                        float(item_a_s_a.b_33), float(item_a_s_a.b_12),
                        float(item_a_s_a.b_13), float(item_a_s_a.b_23))
            l_b_iso.append(b_iso)
            l_beta.append(beta)
        np_b_iso = numpy.array(l_b_iso, dtype=float)
        np_beta = numpy.array(l_beta, dtype=float)
        return np_b_iso, np_beta


    def calc_f_nucl(self, index_hkl):
        dict_crystal = self.get_dictionary()
        dict_in_out = {"index_hkl": index_hkl}
        f_nucl, dder = calc_f_nucl_by_dictionary(dict_crystal, dict_in_out, flag_use_precalculated_data=False)
        return f_nucl


    def calc_refln(self, index_hkl,
                   flag_internal: bool = True):
        """
        Calculate Refln cryspy object where nuclear structure factor is stored.

        Keyword Arguments:
        -----------------
            h, k, l: 1D numpy array of Miller indexes
            flag_internal: a flag to calculate or to use internal objects.
                           It should be True if user call the function.
                           It's True by default.

        Output:
        -------
            refln: object cryspy.Refln

        Example:
        -------
            >>> import numpy as np
            >>> h, k, l = np.array([1,2],dtype=int), np.array([1,0],dtype=int),
                          np.array([1,0],dtype=int)
            >>> refln = crystal.calc_refln(h, k, l)
            >>> print(refln.to_cif())
        """
        # if isinstance(index_h, (float, int)):
        #     index_h = numpy.array([index_h], dtype=float)
        #     index_k = numpy.array([index_k], dtype=float)
        #     index_l = numpy.array([index_l], dtype=float)
        # elif isinstance(index_h, list):
        #     index_h = numpy.array(index_h, dtype=float)
        #     index_k = numpy.array(index_k, dtype=float)
        #     index_l = numpy.array(index_l, dtype=float)

        # index_hkl = numpy.stack([index_h, index_k, index_l], axis=0)
        f_nucl = self.calc_f_nucl(index_hkl)
        res = ReflnL(loop_name=self.data_name)
        res.numpy_index_h = index_hkl[0]
        res.numpy_index_k = index_hkl[1]
        res.numpy_index_l = index_hkl[2]
        res.numpy_f_calc = f_nucl
        if flag_internal:
            res.numpy_to_items()
        return res

    def calc_structure_factor_tensor_ccs(
            self, index_hkl, flag_only_orbital: bool = False, dict_in_out: dict = None):
        """Calculate structure factor tensor in CCS coordinate (X||a*, Z||c)."""
        
        dict_crystal = self.get_dictionary()
        if dict_in_out is None:
            dict_in_out = {"index_hkl": index_hkl, "flag_only_orbital": flag_only_orbital}
        else:
            dict_in_out["index_hkl"] = index_hkl
            dict_in_out["flag_only_orbital"] = flag_only_orbital

        sft_ccs, dder = calc_sft_ccs_by_dictionary(dict_crystal, dict_in_out, flag_use_precalculated_data=False)
        return sft_ccs, dder


    def calc_refln_susceptibility(
            self, index_hkl, flag_internal: bool = True,
            flag_only_orbital: bool = False):
        """
        Calculate susceptibility tensor and moment tensor.

        They are given in Cartesian orthogonal system (x||a*, z||c).

        Keyword Arguments:
        -----------------
            h, k, l: 1D numpy array of Miller indexes

        Output:
        -------
            ReflnSusceptibilityL object of cryspy

        Example:
        -------
            >>> import numpy as np
            >>> h, k, l = np.array([1,2],dtype=int), np.array([1,0],dtype=int),
                          np.array([1,0],dtype=int)
            >>> refln_suscept = crystal.calc_refln_susceptibility(h, k, l)
        """
        
        sft_ccs, dder = self.calc_structure_factor_tensor_ccs(
            index_hkl, flag_only_orbital=flag_only_orbital)

        res = ReflnSusceptibilityL(loop_name=self.data_name)
        res.numpy_index_h = index_hkl[0,:]
        res.numpy_index_k = index_hkl[1,:]
        res.numpy_index_l = index_hkl[2,:]
        res.numpy_chi_11_calc = sft_ccs[0, :]
        res.numpy_chi_12_calc = sft_ccs[1, :]
        res.numpy_chi_13_calc = sft_ccs[2, :]
        res.numpy_chi_21_calc = sft_ccs[3, :]
        res.numpy_chi_22_calc = sft_ccs[4, :]
        res.numpy_chi_23_calc = sft_ccs[5, :]
        res.numpy_chi_31_calc = sft_ccs[6, :]
        res.numpy_chi_32_calc = sft_ccs[7, :]
        res.numpy_chi_33_calc = sft_ccs[8, :]
        if flag_internal:
            res.numpy_to_items()
        return res

    def calc_f_m_perp(self, index_hkl, magnetic_field, dict_in_out: dict = None):
        """Calculate F_M_perpendicular
        """
        sft_ccs, dder_sft_ccs = self.calc_structure_factor_tensor_ccs(index_hkl, dict_in_out=dict_in_out)
        unit_cell_parameters = self.cell.get_unit_cell_parameters()
        eq_ccs, dder_eq_ccs = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters)
        f_m_perp, dder_f_m_perp = calc_f_m_perp_by_sft(
            sft_ccs, magnetic_field, eq_ccs,
            flag_sft_ccs=False,
            flag_magnetic_field=False,
            flag_eq_ccs=False)
        return f_m_perp


    def calc_hkl(self, sthol_min: float = 0., sthovl_max: float = 1.):
        """
        Calculate hkl and multiplicity taking into account the symmetry
        constraints.

        Parameters
        ----------
        sthol_min : float
            minimal sin(theta)/wavelength in inversed angstrems.
        sthovl_max : float
            maximal sin(theta)/wavelength in inversed angstrems.

        Returns
        -------
        h, k, l, mult : numpy.array[float]
            The Miller indeces: h, k, l, and its multiplicity: mult.

        """
        cell = self.cell
        space_group = self.space_group
        res = cell.calc_hkl(space_group, sthol_min, sthovl_max)
        return res


    def calc_hkl_in_range(self, sthol_min: float = 0., sthovl_max: float = 1.):
        """
        Give hkl without taking symmetry into account.

        Parameters
        ----------
        sthol_min : float
            DESCRIPTION.
        sthovl_max : float
            DESCRIPTION.

        Returns
        -------
        res : TYPE
            DESCRIPTION.

        """
        cell = self.cell
        res = cell.calc_hkl_in_range(sthol_min, sthovl_max)
        return res


    def calc_magnetization_ellipsoid(self):
        """
        Magnetization ellipsoids.

        The same coordinate system as U_ij (anisotropic Debye-Waller factor)

        Negtive eigenvalues of ellipsoid are replaced by positive.
        """
        try:
            cell = self.cell
            a_s_m_a = self.atom_site_susceptibility
        except AttributeError:
            return l_res

        l_res = []
        for item in a_s_m_a.items:
            chi_as_u_loc = magnetization_ellipsoid_by_u_ij(cell, item)
            l_res.append(chi_as_u_loc)
        return l_res


    def calc_main_axes_of_magnetization_ellipsoids(self):
        """Susceptibility along the main axes of magnetization ellipsoid.

        Output
        ------
            - l_moments is main axes of ellipsoid in mu_B/T for each atom
            - l_moments_sigma is sigmas for main axes of ellipsoid for each
              atom
            - l_rot_matrix is directions for moments
                for moments[0] direction is rot_matrix[:, 0]
                for moments[1] direction is rot_matrix[:, 1]
                for moments[2] direction is rot_matrix[:, 2]

        The main axes are given in Cartezian coordinate system (x||a*, z||c).
        """
        ll_moments = []
        ll_directions = []
        try:
            cell = self.cell
            a_s_s = self.atom_site_susceptibility
        except AttributeError:
            return ll_moments, ll_directions

        l_moments, l_moments_sigma, l_rot_matrix = \
            a_s_s.calc_main_axes_of_magnetization_ellipsoid(cell)
        return l_moments, l_moments_sigma, l_rot_matrix


    def report_main_axes_of_magnetization_ellipsoids(self):
        """
        Report about main axes of magnetization ellipsoids.

        Make a report about magnetization ellipsoids in string format.
        Calculations are performed by
        get_main_axes_of_magnetization_ellipsoids method
        and calc_magnetization_ellipsoid method of the crystal object.
        """
        # crystal is defined object of cryspy library;
        # type(crystal) is Crystal
        if self.is_attribute("atom_site_susceptibility"):
            a_s_m_a = self.atom_site_susceptibility
        else:
            return ""

        cell = self.cell
        s_out = report_main_axes_of_magnetization_ellipsoids(
            cell, a_s_m_a)

        return s_out


    def calc_magnetic_moments_with_field_loc(self, field_abc):
        """Orientation of magetic moment for atoms at applied magnetic field.

        The input magnetic field should be given in normalized unit cell
        (a/|a|, b/|b|, c/|c|)
        The output magnetic moment are given in normalized unit cell
        (a/|a|, b/|b|, c/|c|)
        
        Output
        ------
        l_lab_out - label of atoms
        l_xyz_out - position of atoms
        l_moment_out - moment of atoms

        """
        np_field = numpy.array(field_abc, dtype=float)
        norm_field = numpy.sqrt(numpy.square(np_field).sum())
        if norm_field == 0.: 
            norm_field = 1.
        np_field = np_field/norm_field
        l_lab_out, l_xyz_out, l_moment_out = [], [], []

        try:
            spgr = self.space_group
            cell = self.cell
            a_s = self.atom_site
            a_s_m_a = self.atom_site_susceptibility
        except AttributeError:
            return None


        m_m_norm = cell.m_m_norm
        m_mt_norm_m_norm_field = numpy.matmul(
            numpy.matmul(m_m_norm.transpose(), m_m_norm), np_field)

        as_p1 = []
        asm_p1 = []
        for _l, _11, _22, _33, _12, _13, _23 in zip(
                a_s_m_a.label, a_s_m_a.chi_11, a_s_m_a.chi_22, a_s_m_a.chi_33,
                a_s_m_a.chi_12, a_s_m_a.chi_13, a_s_m_a.chi_23):
            m_chi = numpy.array([[_11, _12, _13],
                                 [_12, _22, _23],
                                 [_13, _23, _33]], dtype=float)
            x, y, z = a_s[_l].fract_x, a_s[_l].fract_y, a_s[_l].fract_z
            l_out = spgr.calc_rotated_matrix_for_position(m_chi, x, y, z)
            for _i_out, _out in enumerate(l_out):
                _xyz = _out[0]
                _chi = _out[1]
                _moment = numpy.matmul(_chi, m_mt_norm_m_norm_field)
                label_p1 = f"{_l:}_{_i_out+1:}"
                l_lab_out.append(label_p1)
                l_xyz_out.append(_xyz)
                l_moment_out.append(_moment)
                as_p1.append(AtomSite(label=label_p1, fract_x=_xyz[0], fract_y=_xyz[1], fract_z=_xyz[2]))
                asm_p1.append(AtomSiteMoment(
                    label=label_p1,
                    crystalaxis_x=numpy.round(_moment[0], 5),
                    crystalaxis_y=numpy.round(_moment[1], 5),
                    crystalaxis_z=numpy.round(_moment[2], 5)))


        spgr_p1 = SpaceGroup(it_number=1)
        spgr_p1.form_object()
        atom_site_p1 = AtomSiteL()
        atom_site_p1.items = as_p1
        atom_site_moment_p1 = AtomSiteMomentL()
        atom_site_moment_p1.items = asm_p1

        cryspy_p1 = Crystal(space_group = spgr_p1, cell=cell, atom_site=atom_site_p1, atom_site_moment=atom_site_moment_p1)

        return cryspy_p1


    def report(self):
        return self.report_main_axes_of_magnetization_ellipsoids()


    def plots(self):
        return []


    def get_flags_atom_beta(self):
        atom_site = self.atom_site
        if self.is_attribute("atom_site_aniso"):
            atom_site_aniso = self.atom_site_aniso
            l_f = []
            for item_as in atom_site.items:
                if item_as.adp_type == "Bani":
                    item_asa = atom_site_aniso[item_as.label]
                    l_f.append(
                        numpy.array([
                            item_asa.b_11_refinement, item_asa.b_22_refinement, item_asa.b_33_refinement,
                            item_asa.b_12_refinement, item_asa.b_13_refinement, item_asa.b_23_refinement],
                            dtype=bool))
                elif item_as.adp_type == "Uani":
                    item_asa = atom_site_aniso[item_as.label]
                    l_f.append(
                        numpy.array([
                            item_asa.u_11_refinement, item_asa.u_22_refinement, item_asa.u_33_refinement,
                            item_asa.u_12_refinement, item_asa.u_13_refinement, item_asa.u_23_refinement],
                            dtype=bool))
                else:
                    l_f.append(numpy.zeros((6,), dtype=bool))
            res = numpy.stack(l_f, axis=0).transpose()
        else:
            res = numpy.zeros((6, len(atom_site.items)), dtype=bool)
        return res

    def get_dictionary(self):
        """Form dictionary. See documentation moduel CrysPy using Jupyter notebook.
        """
        self.form_object()
        ddict = {}
        space_group, cell, atom_site = None, None, None
        atom_site_susceptibility, atom_electron_configuration = None, None
        atom_site_exchange = None
        atom_type_scat = None
        atom_type = None
        atom_site_aniso = None
        atom_site_scat = None
        atom_site_moment = None
        space_group_symop_magn_centering = None
        space_group_symop_magn_operation = None

        l_obj = take_items_by_class(self, (SpaceGroup,))
        if len(l_obj) > 0:
            space_group = l_obj[0]

        l_obj = take_items_by_class(self, (Cell,))
        if len(l_obj) > 0:
            cell = l_obj[0]

        l_obj = take_items_by_class(self, (AtomSiteL,))
        if len(l_obj) > 0:
            atom_site = l_obj[0]

        l_obj = take_items_by_class(self, (AtomSiteAnisoL,))
        if len(l_obj) > 0:
            atom_site_aniso = l_obj[0]

        l_obj = take_items_by_class(self, (AtomSiteScatL,))
        if len(l_obj) > 0:
            atom_site_scat = l_obj[0]

        l_obj = take_items_by_class(self, (AtomTypeL,))
        if len(l_obj) > 0:
            atom_type = l_obj[0]

        l_obj = take_items_by_class(self, (AtomTypeScatL,))
        if len(l_obj) > 0:
            atom_type_scat = l_obj[0]

        l_obj = take_items_by_class(self, (AtomSiteSusceptibilityL,))
        if len(l_obj) > 0:
            atom_site_susceptibility = l_obj[0]

        l_obj = take_items_by_class(self, (AtomSiteExchangeL,))
        if len(l_obj) > 0:
            atom_site_exchange = l_obj[0]
            

        l_obj = take_items_by_class(self, (AtomElectronConfigurationL,))
        if len(l_obj) > 0:
            atom_electron_configuration = l_obj[0]

        l_obj = take_items_by_class(self, (AtomSiteMomentL,))
        if len(l_obj) > 0:
            atom_site_moment = l_obj[0]

        l_obj = take_items_by_class(self, (SpaceGroupSymopMagnCenteringL,))
        if len(l_obj) > 0:
            space_group_symop_magn_centering = l_obj[0]

        l_obj = take_items_by_class(self, (SpaceGroupSymopMagnOperationL,))
        if len(l_obj) > 0:
            space_group_symop_magn_operation = l_obj[0]

        ddict["name"] = self.data_name
        ddict["type_name"] = self.get_name()
        ind_mag = None
        rad = numpy.pi/180.
        if space_group is not None:
            full_symm_elems = space_group.full_space_group_symop.get_symm_elems()
            ddict["full_symm_elems"] = full_symm_elems
            reduced_symm_elems = space_group.reduced_space_group_symop.get_symm_elems()
            ddict["reduced_symm_elems"] = reduced_symm_elems
            ddict["centrosymmetry"] = space_group.centrosymmetry
            if ddict["centrosymmetry"]:
                p_centr = space_group.pcentr
                lcm = numpy.lcm.reduce([fr.denominator for fr in p_centr])
                vals = [int(fr.numerator * lcm / fr.denominator) for fr in p_centr]
                vals.append(lcm)
                ddict["centrosymmetry_position"] = numpy.array(vals, dtype=int)

            shift = space_group.shift
            l_vals = []
            for ind in range(len(shift)):
                lcm = numpy.lcm.reduce([fr.denominator for fr in shift[ind]])
                vals = [int(fr.numerator * lcm / fr.denominator) for fr in shift[ind]]
                vals.append(lcm)
                l_vals.append(vals)
            ddict["translation_elems"] = numpy.array(l_vals, dtype=int).transpose()

        if space_group_symop_magn_operation is not None:
            if space_group_symop_magn_centering is not None:
                se_sgsmo = self.space_group_symop_magn_operation.get_sym_elems()
                se_sgsmc = self.space_group_symop_magn_centering.get_sym_elems()
                full_mcif_elems = calc_full_mag_elems(se_sgsmo, se_sgsmc)
                ddict["full_mcif_elems"] = full_mcif_elems

        if cell is not None:
            if cell.is_attribute("type_cell"):
                type_cell, it_coordinate_system_code = cell.type_cell, cell.it_coordinate_system_code
            elif "full_mcif_elems" in ddict.keys(): 
                type_cell, it_coordinate_system_code = define_bravais_type_by_symm_elems(full_mcif_elems)
                cell.type_cell = type_cell
                cell.it_coordinate_system_code = it_coordinate_system_code
                cell.apply_constraints()

            unit_cell_parameters = cell.get_unit_cell_parameters()
            
            sc_uc, v_uc = calc_sc_v_unit_cell_parameters(type_cell, it_coordinate_system_code)
            ddict["sc_uc"] = sc_uc
            ddict["v_uc"] = v_uc
            
            ddict["unit_cell_parameters"] = numpy.dot(sc_uc, unit_cell_parameters)+v_uc
            ddict["flags_unit_cell_parameters"] = cell.get_flags_unit_cell_parameters()

        if atom_site is not None:
            atom_label = numpy.array(atom_site.label, dtype=str)
            atom_occupancy = numpy.array(atom_site.occupancy, dtype=float)
            if atom_site.is_attribute("multiplicity"):
                atom_multiplicity = numpy.array(atom_site.multiplicity, dtype=int)
                ddict["atom_multiplicity"] = atom_multiplicity
            atom_fract_xyz = numpy.array([atom_site.fract_x,
                                          atom_site.fract_y,
                                          atom_site.fract_z], dtype=float)
            atom_fract_xyz = numpy.mod(atom_fract_xyz, 1.)

            atom_type_symbol = numpy.array(atom_site.type_symbol, dtype=str)
            table_sthovl = numpy.linspace(0, 2, 501) # fro0 to 2 Å**-1
            d_dispersion = DATABASE["Dispersion"]
            table_wavelength = d_dispersion["table_wavelength"]
            l_table_atom_scattering_amplitude = []
            l_table_atom_dispersion = []
            flag_atom_scattering_amplitude = True
            for type_symbol in atom_type_symbol:
                try:
                    jl = calc_jl_for_ion(table_sthovl, type_symbol)
                    scattering_amplitude = jl[:,0]
                except UserWarning:
                    try:
                        scattering_amplitude = calc_scattering_amplitude_tabulated(type_symbol, table_sthovl)[0]
                    except KeyError:
                        flag_atom_scattering_amplitude = False
                if flag_atom_scattering_amplitude:
                    l_table_atom_scattering_amplitude.append(scattering_amplitude)

                type_atom = get_atom_name_ion_charge_shell(type_symbol)[0]
                try:
                    l_table_atom_dispersion.append(d_dispersion[type_atom.capitalize()])
                except KeyError:
                    l_table_atom_dispersion.append(numpy.zeros_like(table_wavelength))
            if flag_atom_scattering_amplitude:
                ddict["table_sthovl"] = table_sthovl
                ddict["table_atom_scattering_amplitude"] = numpy.stack(l_table_atom_scattering_amplitude, axis=0)
            ddict["table_wavelength"] = table_wavelength
            ddict["table_atom_dispersion"] = numpy.stack(l_table_atom_dispersion, axis=0)

            atom_b_scat = numpy.array(atom_site.scat_length_neutron, dtype=complex)

            b_d  = 3*10**6*numpy.ones(atom_fract_xyz.shape[1:], dtype=int) #10**6 it is precision
            b_elem = (atom_fract_xyz*b_d[na, :]).astype(int)
            gcd = numpy.gcd.reduce(numpy.gcd(b_elem, b_d[na,:]))

            atom_symm_elems = numpy.concatenate([
                numpy.floor_divide(b_elem, gcd[na, :]),
                numpy.floor_divide(b_d, gcd)[na, :],
                numpy.zeros((9, ) + atom_fract_xyz.shape[1:], dtype=int)],
                axis=0)


            if "full_mcif_elems" in ddict.keys():
                elems_fs = ddict["full_mcif_elems"]
            elif "full_symm_elems" in ddict.keys():
                elems_fs = ddict["full_symm_elems"]
            else:
                raise AttributeError("Symmetry elements are not found.")

            flag_2d = calc_symm_flags(elems_fs[:, :, na], atom_symm_elems[:, na, :])


            l_sc_fract, l_sc_b = [], []
            for ind in range(atom_label.size):
                symm_elems = elems_fs[:, flag_2d[:, ind]]
                sc_fract, sc_b = calc_sc_fract_sc_b(symm_elems, atom_fract_xyz[:, ind])
                l_sc_fract.append(sc_fract)
                l_sc_b.append(sc_b)
            atom_site_sc_fract = numpy.stack(l_sc_fract, axis=-1)

            ddict["atom_site_sc_fract"] = atom_site_sc_fract 
            ddict["atom_site_sc_b"] = numpy.stack(l_sc_b, axis=-1)

            flags_atom_fract_xyz = atom_site.get_flags_atom_fract_xyz()
            flags_atom_fract_xyz_constr = calc_m_v(atom_site_sc_fract, flags_atom_fract_xyz, flag_m=False, flag_v=False)[0].astype(bool)

            ddict["atom_label"] = atom_label
            ddict["atom_fract_xyz"] = atom_fract_xyz
            ddict["flags_atom_fract_xyz"] = flags_atom_fract_xyz_constr 
            ddict["atom_scat_length_neutron"] = atom_b_scat
            ddict["atom_occupancy"] = atom_occupancy
            ddict["flags_atom_occupancy"] = atom_site.get_flags_atom_occupancy()
            ddict["flags_atom_b_iso"] = atom_site.get_flags_atom_b_iso()
            ddict["atom_symm_elems_flag"] = flag_2d

        if atom_site_susceptibility is not None:
            atom_para_label = numpy.array(atom_site_susceptibility.label, dtype=str)
            atom_para_chi_type = numpy.array(atom_site_susceptibility.chi_type, dtype=str)
            atom_para_susceptibility = numpy.array([
                atom_site_susceptibility.chi_11, atom_site_susceptibility.chi_22,
                atom_site_susceptibility.chi_33, atom_site_susceptibility.chi_12,
                atom_site_susceptibility.chi_13, atom_site_susceptibility.chi_23],
                dtype=float)
            ddict["atom_para_label"] = atom_para_label
            ddict["atom_para_susceptibility"] = atom_para_susceptibility
            ddict["flags_atom_para_susceptibility"] = atom_site_susceptibility.get_flags_susceptibility()            
            if "atom_label" in ddict.keys():
                flag_2d = numpy.expand_dims(ddict["atom_label"], axis=1) == numpy.expand_dims(atom_para_label, axis=0)
                args = numpy.argwhere(flag_2d)
                atom_para_index = args[args[:,1].argsort()][:,0]
                ddict["atom_para_index"] = atom_para_index

                if "full_mcif_elems" in ddict.keys():
                    elems_fs = ddict["full_mcif_elems"]
                elif "full_symm_elems" in ddict.keys():
                    elems_fs = ddict["full_symm_elems"]
                else:
                    raise AttributeError("Symmetry elements are not found.")
                flag_2d = ddict["atom_symm_elems_flag"][:, atom_para_index]

                l_sc_chi = []
                for ind in range(atom_para_label.size):
                    if (atom_para_chi_type[ind]).lower() == "ciso":
                        sc_chi = numpy.array([
                            [1., 0., 0., 0., 0., 0.],
                            [1., 0., 0., 0., 0., 0.],
                            [1., 0., 0., 0., 0., 0.],
                            [0., 0., 0., 0., 0., 0.],
                            [0., 0., 0., 0., 0., 0.],
                            [0., 0., 0., 0., 0., 0.]], dtype=float)
                    else:
                        sc_chi = calc_sc_chi(
                            elems_fs[:, flag_2d[:, ind]],
                            unit_cell_parameters=unit_cell_parameters, flag_unit_cell_parameters=False)[0]
                    l_sc_chi.append(sc_chi)
                atom_para_sc_chi = numpy.stack(l_sc_chi, axis=-1)
                ddict["atom_para_sc_chi"] = atom_para_sc_chi
        
        if atom_site_exchange is not None:
            atom_exchange_label = numpy.array(atom_site_exchange.label, dtype=str)
            atom_exchange_type = numpy.array(atom_site_exchange.j_type, dtype=str)
            atom_exchange_tensor = numpy.array([
                atom_site_exchange.j_11, 
                atom_site_exchange.j_22, 
                atom_site_exchange.j_33, 
                atom_site_exchange.j_12, 
                atom_site_exchange.j_13, 
                atom_site_exchange.j_23, 
                ], dtype=float)
            flags_atom_exchange_tensor = numpy.array([
                atom_site_exchange.j_11_refinement,
                atom_site_exchange.j_22_refinement,
                atom_site_exchange.j_33_refinement,
                atom_site_exchange.j_12_refinement,
                atom_site_exchange.j_13_refinement,
                atom_site_exchange.j_23_refinement,
                ], dtype=bool)
            ddict["atom_exchange_label"] = atom_exchange_label
            ddict["atom_exchange_type"] = atom_exchange_type
            ddict["atom_exchange_tensor"] = atom_exchange_tensor
            ddict["flags_atom_exchange_tensor"] = flags_atom_exchange_tensor
            if "atom_label" in ddict.keys():

                if "full_mcif_elems" in ddict.keys():
                    elems_fs = ddict["full_mcif_elems"]
                elif "full_symm_elems" in ddict.keys():
                    elems_fs = ddict["full_symm_elems"]
                else:
                    raise AttributeError("Symmetry elements are not found.")

                flag_2d = numpy.expand_dims(ddict["atom_label"], axis=1) == numpy.expand_dims(atom_exchange_label, axis=0)
                args = numpy.argwhere(flag_2d)
                atom_exchange_index = args[args[:,1].argsort()][:,0]
                ddict["atom_exchange_index"] = atom_exchange_index

                l_sc_exchange = []
                flag_2d = ddict["atom_symm_elems_flag"][:, atom_exchange_index]
                for ind in range(atom_exchange_label.size):
                    if (atom_exchange_type[ind]).lower() == "jiso":
                        sc_exchange = numpy.array([
                            [1., 0., 0., 0., 0., 0.],
                            [1., 0., 0., 0., 0., 0.],
                            [1., 0., 0., 0., 0., 0.],
                            [0., 0., 0., 0., 0., 0.],
                            [0., 0., 0., 0., 0., 0.],
                            [0., 0., 0., 0., 0., 0.]], dtype=float)
                    else:
                        sc_exchange = calc_sc_chi(
                            elems_fs[:, flag_2d[:, ind]],
                            unit_cell_parameters=unit_cell_parameters, flag_unit_cell_parameters=False)[0]
                    l_sc_exchange.append(sc_exchange)
                atom_sc_exchange = numpy.stack(l_sc_exchange, axis=-1)
                ddict["atom_sc_exchange"] = atom_sc_exchange

        if atom_site_moment is not None:
            atom_ordered_label = numpy.array(atom_site_moment.label, dtype=str)
            atom_ordered_moment_crystalaxis_xyz = numpy.array([
                atom_site_moment.crystalaxis_x, 
                atom_site_moment.crystalaxis_y, 
                atom_site_moment.crystalaxis_z], dtype=float)
            flags_atom_ordered_moment_crystalaxis_xyz = numpy.array([
                atom_site_moment.crystalaxis_x_refinement, 
                atom_site_moment.crystalaxis_y_refinement, 
                atom_site_moment.crystalaxis_z_refinement], dtype=bool)
            ddict["atom_ordered_label"] = atom_ordered_label
            ddict["atom_ordered_moment_crystalaxis_xyz"] = atom_ordered_moment_crystalaxis_xyz
            ddict["flags_atom_ordered_moment_crystalaxis_xyz"] = flags_atom_ordered_moment_crystalaxis_xyz
            if "atom_label" in ddict.keys():
                flag_2d = numpy.expand_dims(ddict["atom_label"], axis=1) == numpy.expand_dims(atom_ordered_label, axis=0)
                args = numpy.argwhere(flag_2d)
                atom_ordered_index = args[args[:,1].argsort()][:,0]
                ddict["atom_ordered_index"] = atom_ordered_index

        if isinstance(self, Crystal):
            # FIXME: beta should be defined in atom_site_aniso_beta, b_iso for every atom
            atom_b_iso, atom_beta = self.calc_b_iso_beta()
            ddict["atom_b_iso"] = atom_b_iso
            ddict["atom_beta"] = atom_beta.transpose()
            ddict["flags_atom_beta"] = self.get_flags_atom_beta()

        if atom_site_scat is not None:
            mag_atom_label = numpy.array(atom_site_scat.label, dtype=str)
            mag_atom_lande_factor = numpy.array(atom_site_scat.lande, dtype=float)
            mag_atom_kappa = numpy.array(atom_site_scat.kappa, dtype=float)
            flags_mag_atom_lande_factor = atom_site_scat.get_flags_lande()
            flags_mag_atom_kappa = atom_site_scat.get_flags_kappa()

            ddict["mag_atom_label"] = mag_atom_label
            ddict["mag_atom_lande_factor"] = mag_atom_lande_factor
            ddict["mag_atom_kappa"] = mag_atom_kappa
            ddict["flags_mag_atom_lande_factor"] = flags_mag_atom_lande_factor
            ddict["flags_mag_atom_kappa"] = flags_mag_atom_kappa
            # l_ats = atom_site_scat.atom_type_scat
            if "atom_label" in ddict.keys():
                flag_2d = numpy.expand_dims(ddict["atom_label"], axis=1) == numpy.expand_dims(mag_atom_label, axis=0)
                args = numpy.argwhere(flag_2d)
                mag_atom_index = args[args[:,1].argsort()][:,0]
                ddict["mag_atom_index"] = mag_atom_index

                if atom_site_scat.is_attribute("type_symbol"):
                    mag_type_symbol = numpy.array(atom_site_scat.type_symbol, dtype=str)
                else:
                    as_tsymbol = numpy.array(atom_site.type_symbol, dtype=str)
                    mag_type_symbol = as_tsymbol[mag_atom_index] 

                j0_parameters, j2_parameters = get_j0_j2_parameters(mag_type_symbol)
                ddict["mag_atom_j0_parameters"] = j0_parameters
                ddict["mag_atom_j2_parameters"] = j2_parameters
            
            if "atom_para_label" in ddict.keys():
                flag_2d = numpy.expand_dims(ddict["atom_para_label"], axis=0) == numpy.expand_dims(mag_atom_label, axis=1)
                args = numpy.argwhere(flag_2d)
                mag_atom_para_index = args[args[:,1].argsort()][:,0]
                ddict["mag_atom_para_index"] = mag_atom_para_index

            if "atom_ordered_label" in ddict.keys():
                flag_2d = numpy.expand_dims(ddict["atom_ordered_label"], axis=0) == numpy.expand_dims(mag_atom_label, axis=1)
                args = numpy.argwhere(flag_2d)
                mag_atom_ordered_index = args[args[:,1].argsort()][:,0]
                ddict["mag_atom_ordered_index"] = mag_atom_ordered_index

        if atom_type is not None:
            pass

        if atom_type_scat is not None:
            pass

        if atom_site_aniso is not None:
            atom_site_aniso_label = numpy.array(atom_site_aniso.label, dtype=str)
            flag_2d = numpy.expand_dims(ddict["atom_label"], axis=1) == numpy.expand_dims(atom_site_aniso_label, axis=0)
            args = numpy.argwhere(flag_2d)
            atom_site_aniso_index = args[args[:,1].argsort()][:,0]
            ddict["atom_site_aniso_index"] = atom_site_aniso_index
            #FIXME: add transformation U_ani and others to beta

            if "full_mcif_elems" in ddict.keys():
                elems_fs = ddict["full_mcif_elems"]
            elif "full_symm_elems" in ddict.keys():
                elems_fs = ddict["full_symm_elems"]
            else:
                raise AttributeError("Symmetry elements are not found.")

            flag_2d = ddict["atom_symm_elems_flag"][:, atom_site_aniso_index]
            l_sc_beta = []
            atom_site_aniso_label = numpy.array(atom_site_aniso.label, dtype=str)

            for ind in range(atom_site_aniso_label.size):
                symm_elems = elems_fs[:, flag_2d[:, ind]]
                sc_beta = calc_sc_beta(symm_elems)
                l_sc_beta.append(sc_beta)
            ddict["atom_site_aniso_sc_beta"] = numpy.stack(l_sc_beta, axis=-1)


        if atom_electron_configuration is not None:
            d_slater_orbitals = DATABASE["Orbitals by radial Slater functions"]
            for aec_item in atom_electron_configuration.items:
                aec_label = aec_item.label
                at_symbol = atom_site[aec_label].type_symbol
                element_symbol, charge, isotope_number = get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol(at_symbol)
                core_shells_populations = aec_item.get_core_shells_populations()
                l_population = []
                l_n = []
                l_zeta = []
                l_coeff = []
                for sh_pop in core_shells_populations:
                    shell, population = sh_pop
                    n, dzeta, coeff = d_slater_orbitals[(element_symbol, shell)]
                    l_population.append(population)
                    l_n.append(n)
                    l_zeta.append(dzeta)
                    l_coeff.append(coeff)
                if len(l_population) > 0:
                    dict_shell = {"core_population": l_population,
                        "core_n": l_n,
                        "core_zeta": l_zeta,
                        "core_coeff": l_coeff}
                    ddict[f"shell_{aec_label:}"] = dict_shell
        return ddict

    def take_parameters_from_dictionary(self, ddict_crystal, l_parameter_name: list=None, l_sigma: list=None):
        keys = ddict_crystal.keys()
        if l_parameter_name is None:
            l_parameter_name = []
            l_sigma = []

        if "atom_para_susceptibility" in keys:
            mag_atom_susceptibility = (ddict_crystal["atom_para_sc_chi"]  * ddict_crystal["atom_para_susceptibility"][na, :, :]).sum(axis=1)
            for i_item, item_ass in enumerate(self.atom_site_susceptibility.items):
                item_ass.chi_11 = mag_atom_susceptibility[0, i_item]
                item_ass.chi_22 = mag_atom_susceptibility[1, i_item]
                item_ass.chi_33 = mag_atom_susceptibility[2, i_item]
                item_ass.chi_12 = mag_atom_susceptibility[3, i_item]
                item_ass.chi_13 = mag_atom_susceptibility[4, i_item]
                item_ass.chi_23 = mag_atom_susceptibility[5, i_item]

        if "atom_exchange_tensor" in keys:
            j_tensor = ddict_crystal["atom_exchange_tensor"]
            atom_sc_exchange = ddict_crystal["atom_sc_exchange"]
            j_tensor = numpy.sum(numpy.expand_dims(j_tensor, axis=0)*atom_sc_exchange, axis=1)
            for i_item, item_ase in enumerate(self.atom_site_exchange.items):
                item_ase.j_11 = j_tensor[0, i_item]
                item_ase.j_22 = j_tensor[1, i_item]
                item_ase.j_33 = j_tensor[2, i_item]
                item_ase.j_12 = j_tensor[3, i_item]
                item_ase.j_13 = j_tensor[4, i_item]
                item_ase.j_23 = j_tensor[5, i_item]

        if "atom_ordered_moment_crystalaxis_xyz" in keys:
            hh = ddict_crystal["atom_ordered_moment_crystalaxis_xyz"]
            for i_item, item in enumerate(self.atom_site_moment.items):
                item.crystalaxis_x = hh[0, i_item]
                item.crystalaxis_y = hh[1, i_item]
                item.crystalaxis_z = hh[2, i_item]

        if "mag_atom_lande_factor" in keys:
            hh = ddict_crystal["mag_atom_lande_factor"]
            for i_item, item in enumerate(self.atom_site_scat.items):
                item.lande = hh[i_item]

        if "mag_atom_kappa" in keys:
            hh = ddict_crystal["mag_atom_kappa"]
            for i_item, item in enumerate(self.atom_site_scat.items):
                item.kappa = hh[i_item]

        if "atom_fract_xyz" in keys:
            hh = ddict_crystal["atom_fract_xyz"]
            atom_site_sc_fract = ddict_crystal["atom_site_sc_fract"] 
            atom_site_sc_b = ddict_crystal["atom_site_sc_b"] 
            hh = calc_m_v(atom_site_sc_fract, numpy.mod(hh, 1), flag_m=False, flag_v=False)[0] + atom_site_sc_b

            for i_item, item in enumerate(self.atom_site.items):
                item.fract_x = numpy.round(numpy.mod(hh[0, i_item], 1.), decimals=6)
                item.fract_y = numpy.round(numpy.mod(hh[1, i_item], 1.), decimals=6)
                item.fract_z = numpy.round(numpy.mod(hh[2, i_item], 1.), decimals=6)

        if "atom_occupancy" in keys:
            hh = ddict_crystal["atom_occupancy"]
            for i_item, item in enumerate(self.atom_site.items):
                item.occupancy = numpy.round(hh[i_item], decimals=6)

        if "atom_multiplicity" in keys:
            hh = ddict_crystal["atom_multiplicity"]
            for i_item, item in enumerate(self.atom_site.items):
                item.multiplicity = hh[i_item]

        if ("atom_b_iso" in keys):
            atom_b_iso = ddict_crystal["atom_b_iso"]
            if "atom_beta" in keys:
                atom_beta = ddict_crystal["atom_beta"]

                if "atom_site_aniso_sc_beta" in keys:
                    atom_site_aniso_sc_beta = ddict_crystal["atom_site_aniso_sc_beta"]
                    atom_site_aniso_index = ddict_crystal["atom_site_aniso_index"]
                    atom_sc_beta = numpy.zeros((6,)+atom_beta.shape, dtype=float)
                    atom_sc_beta[:, :, atom_site_aniso_index] = atom_site_aniso_sc_beta
                    atom_beta = (atom_sc_beta*numpy.expand_dims(atom_beta, axis=0)).sum(axis=1)

            else:
                atom_beta = numpy.zeros((6, )+ atom_b_iso.shape, dtype=float)
            unit_cell_parameters = ddict_crystal["unit_cell_parameters"]
            if not(self.atom_site.is_attribute("adp_type")):
                for item in self.atom_site:
                    item.adp_type = "Biso"
                    if not(item.is_attribute("b_iso_or_equiv")):
                        item.b_iso_or_equiv = 0.
            atom_adp_type = numpy.array(self.atom_site.adp_type, dtype=str)
            atom_iso_param, atom_aniso_param = calc_param_iso_aniso_by_b_iso_beta(unit_cell_parameters, atom_adp_type, atom_b_iso, atom_beta)


            if self.is_attribute("atom_site"):
                atom_site = self.atom_site
                for i_item, item in enumerate(self.atom_site.items):
                    if item.adp_type in ("Uiso", "Uovl", "Umpe"):
                        item.u_iso_or_equiv = atom_iso_param[i_item]
                    elif item.adp_type in ("Biso", "Bovl"):
                        item.b_iso_or_equiv = atom_iso_param[i_item]

                if self.is_attribute("atom_site_aniso"):
                    atom_site_aniso_index = ddict_crystal["atom_site_aniso_index"]

                    for item, index_as in zip(self.atom_site_aniso.items, atom_site_aniso_index):
                        aniso_param = atom_aniso_param[:, index_as]
                        item_as = self.atom_site.items[index_as]
                        if item_as.adp_type in ("Uani", ):
                            item.u_11 = aniso_param[0]
                            item.u_22 = aniso_param[1]
                            item.u_33 = aniso_param[2]
                            item.u_12 = aniso_param[3]
                            item.u_13 = aniso_param[4]
                            item.u_23 = aniso_param[5]
                        elif item_as.adp_type in ("Bani", ):
                            item.b_11 = aniso_param[0]
                            item.b_22 = aniso_param[1]
                            item.b_33 = aniso_param[2]
                            item.b_12 = aniso_param[3]
                            item.b_13 = aniso_param[4]
                            item.b_23 = aniso_param[5]

        if "unit_cell_parameters" in keys:
            hh = ddict_crystal["unit_cell_parameters"]
            sc_uc = ddict_crystal["sc_uc"]
            v_uc = ddict_crystal["v_uc"]
            hhh = numpy.dot(sc_uc, hh) + v_uc
            cell = self.cell
            cell.length_a = hhh[0]
            cell.length_b = hhh[1]
            cell.length_c = hhh[2]
            cell.angle_alpha = numpy.round(hhh[3]*180./numpy.pi, decimals=5)
            cell.angle_beta = numpy.round(hhh[4]*180./numpy.pi, decimals=5)
            cell.angle_gamma = numpy.round(hhh[5]*180./numpy.pi, decimals=5)

        for name, sigma in zip(l_parameter_name, l_sigma):
            parameter_label, ind_s = name
            if "atom_para_susceptibility" in parameter_label:
                ind_chi, ind_a = ind_s[0], ind_s[1]
                item_ass = self.atom_site_susceptibility.items[ind_a]
                if ind_chi == 0:
                    item_ass.chi_11_sigma = float(sigma)
                elif ind_chi == 1:
                    item_ass.chi_22_sigma = float(sigma)
                elif ind_chi == 2:
                    item_ass.chi_33_sigma = float(sigma)
                elif ind_chi == 3:
                    item_ass.chi_12_sigma = float(sigma)
                elif ind_chi == 4:
                    item_ass.chi_13_sigma = float(sigma)
                elif ind_chi == 5:
                    item_ass.chi_23_sigma = float(sigma)
            
            if "atom_exchange_tensor" in parameter_label:
                ind_j, ind_a = ind_s[0], ind_s[1]
                item = self.atom_site_exchange.items[ind_a]
                if ind_j == 0:
                    item.j_11_sigma = float(sigma)
                elif ind_j == 1:
                    item.j_22_sigma = float(sigma)
                elif ind_j == 2:
                    item.j_33_sigma = float(sigma)
                elif ind_j == 3:
                    item.j_12_sigma = float(sigma)
                elif ind_j == 4:
                    item.j_13_sigma = float(sigma)
                elif ind_j == 5:
                    item.j_23_sigma = float(sigma)
            
            if "atom_fract_xyz" in parameter_label:
                ind_p, ind_a = ind_s[0], ind_s[1]
                item = self.atom_site.items[ind_a]
                if ind_p == 0:
                    item.fract_x_sigma = float(sigma)
                elif ind_p == 1:
                    item.fract_y_sigma = float(sigma)
                elif ind_p == 2:
                    item.fract_z_sigma = float(sigma)
            if "atom_occupancy" in parameter_label:
                ind_a = ind_s[0]
                item = self.atom_site.items[ind_a]
                item.occupancy_sigma = float(sigma)

            if "atom_b_iso" in parameter_label:
                ind_a = ind_s[0]
                item = self.atom_site.items[ind_a]
                if item.adp_type in ("Uiso", "Uovl", "Umpe"):
                    item.u_iso_or_equiv_sigma = float(sigma)/float(8.*numpy.square(numpy.pi))
                elif item.adp_type in ("Biso", "Bovl", "Bmpe"):
                    item.b_iso_or_equiv_sigma = float(sigma)

            if "atom_beta" in parameter_label:
                ind_p, ind_a = ind_s[0], ind_s[1]
                item = self.atom_site_aniso[self.atom_site.items[ind_a].label]
                if self.atom_site[item.label].adp_type in ("Bani", ):
                    if ind_p == 0:
                        item.b_11_sigma = float(sigma)
                    elif ind_p == 1:
                        item.b_22_sigma = float(sigma)
                    elif ind_p == 2:
                        item.b_33_sigma = float(sigma)
                    elif ind_p == 3:
                        item.b_12_sigma = float(sigma)
                    elif ind_p == 4:
                        item.b_13_sigma = float(sigma)
                    elif ind_p == 5:
                        item.b_23_sigma = float(sigma)

                if self.atom_site[item.label].adp_type in ("Uani", ):
                    reciprocal_unit_cell_parameters = calc_reciprocal_by_unit_cell_parameters(unit_cell_parameters, flag_unit_cell_parameters=False)[0]
                    beta_hh = numpy.zeros((6,1), dtype=float)
                    coeff_hh = calc_u_ij_by_beta(beta_hh, reciprocal_unit_cell_parameters, flag_beta=True)[1]["beta"]
                    if ind_p == 0:
                        item.u_11_sigma = float(sigma*coeff_hh[0])
                    elif ind_p == 1:
                        item.u_22_sigma = float(sigma*coeff_hh[1])
                    elif ind_p == 2:
                        item.u_33_sigma = float(sigma*coeff_hh[2])
                    elif ind_p == 3:
                        item.u_12_sigma = float(sigma*coeff_hh[3])
                    elif ind_p == 4:
                        item.u_13_sigma = float(sigma*coeff_hh[4])
                    elif ind_p == 5:
                        item.u_23_sigma = float(sigma*coeff_hh[5])


            if "atom_ordered_moment_crystalaxis_xyz" in parameter_label:
                ind_p, ind_a = ind_s[0], ind_s[1]
                item = self.atom_site_moment.items[ind_a]
                if ind_p == 0:
                    item.crystalaxis_x_sigma = float(sigma)
                elif ind_p == 1:
                    item.crystalaxis_y_sigma = float(sigma)
                elif ind_p == 2:
                    item.crystalaxis_z_sigma = float(sigma)

            if "mag_atom_kappa" in parameter_label:
                ind_a = ind_s[0]
                item = self.atom_site_scat.items[ind_a]
                item.kappa_sigma = float(sigma)

            if "mag_atom_lande_factor" in parameter_label:
                ind_a = ind_s[0]
                item = self.atom_site_scat.items[ind_a]
                item.lande_sigma = float(sigma)

            if "unit_cell_parameters" in parameter_label:
                ind_p = ind_s[0]
                item = self.cell
                if ind_p == 0:
                    item.length_a_sigma = float(sigma)
                elif ind_p == 1:
                    item.length_b_sigma = float(sigma)
                elif ind_p == 2:
                    item.length_c_sigma = float(sigma)
                elif ind_p == 3:
                    item.angle_alpha_sigma = float(sigma)*180./numpy.pi
                elif ind_p == 4:
                    item.angle_beta_sigma = float(sigma)*180./numpy.pi
                elif ind_p == 5:
                    item.angle_gamma_sigma = float(sigma)*180./numpy.pi



def calc_sc_v_unit_cell_parameters(type_cell: str, it_coordinate_system_code: str):
    """
       Calculate symmetry constraints on unit_cell_parameters as
       unit_cell_parameters = sc_uc * unit_cell_parameters + v_uc
    """
    sc_uc = numpy.eye(6, dtype=float)
    v_uc = numpy.zeros((6,), dtype=float)

    if type_cell == "aP":
        pass
    elif type_cell.startswith("m"):
        sc_uc[3, 3] = 0.
        sc_uc[5, 5] = 0.
        v_uc[3] = numpy.pi/2.
        v_uc[5] = numpy.pi/2.
    elif type_cell.startswith("o"):
        sc_uc[3, 3] = 0.
        sc_uc[4, 4] = 0.
        sc_uc[5, 5] = 0.
        v_uc[3] = numpy.pi/2.
        v_uc[4] = numpy.pi/2.
        v_uc[5] = numpy.pi/2.
    elif ((type_cell.startswith("t"))):  # FIXME: check  | (type_cell == "hP")
        sc_uc[0, 0], sc_uc[0, 1] = 1., 0.
        sc_uc[1, 0], sc_uc[1, 1] = 1., 0.
        sc_uc[3, 3] = 0.
        sc_uc[4, 4] = 0.
        sc_uc[5, 5] = 0.
        v_uc[3] = numpy.pi/2.
        v_uc[4] = numpy.pi/2.
        v_uc[5] = numpy.pi/2.
    elif ((type_cell.startswith("hP"))):
        if it_coordinate_system_code.lower() == "h":
            sc_uc[0, 0], sc_uc[0, 1] = 1., 0.
            sc_uc[1, 0], sc_uc[1, 1] = 1., 0.
            sc_uc[3, 3] = 0.
            sc_uc[4, 4] = 0.
            sc_uc[5, 5] = 0.
            v_uc[3] = numpy.pi/2.
            v_uc[4] = numpy.pi/2.
            v_uc[5] = numpy.pi*2./3.
        else:
            sc_uc[0, 0], sc_uc[0, 1], sc_uc[0, 2] = 1., 0., 0.
            sc_uc[1, 0], sc_uc[1, 1], sc_uc[1, 2] = 1., 0., 0.
            sc_uc[2, 0], sc_uc[2, 1], sc_uc[2, 2] = 1., 0., 0.
            sc_uc[3, 3], sc_uc[3, 4], sc_uc[3, 5] = 1., 0., 0.
            sc_uc[4, 3], sc_uc[4, 4], sc_uc[4, 5] = 1., 0., 0.
            sc_uc[5, 3], sc_uc[5, 4], sc_uc[5, 5] = 1., 0., 0.
    elif (type_cell == "hR"):
        if it_coordinate_system_code.lower() == "h":
            sc_uc[0, 0], sc_uc[0, 1] = 1., 0.
            sc_uc[1, 0], sc_uc[1, 1] = 1., 0.
            sc_uc[3, 3] = 0.
            sc_uc[4, 4] = 0.
            sc_uc[5, 5] = 0.
            v_uc[3] = numpy.pi/2.
            v_uc[4] = numpy.pi/2.
            v_uc[5] = numpy.pi*2./3.
        else:
            sc_uc[0, 0], sc_uc[0, 1], sc_uc[0, 2] = 1., 0., 0.
            sc_uc[1, 0], sc_uc[1, 1], sc_uc[1, 2] = 1., 0., 0.
            sc_uc[2, 0], sc_uc[2, 1], sc_uc[2, 2] = 1., 0., 0.
            sc_uc[3, 3], sc_uc[3, 4], sc_uc[3, 5] = 1., 0., 0.
            sc_uc[4, 3], sc_uc[4, 4], sc_uc[4, 5] = 1., 0., 0.
            sc_uc[5, 3], sc_uc[5, 4], sc_uc[5, 5] = 1., 0., 0.
    elif type_cell.startswith("c"):
        sc_uc[0, 0], sc_uc[0, 1], sc_uc[0, 2] = 1., 0., 0.
        sc_uc[1, 0], sc_uc[1, 1], sc_uc[1, 2] = 1., 0., 0.
        sc_uc[2, 0], sc_uc[2, 1], sc_uc[2, 2] = 1., 0., 0.
        sc_uc[3, 3] = 0.
        sc_uc[4, 4] = 0.
        sc_uc[5, 5] = 0.
        v_uc[3] = numpy.pi/2.
        v_uc[4] = numpy.pi/2.
        v_uc[5] = numpy.pi/2.
    
    return sc_uc, v_uc


import numpy
from cryspy.A_functions_base.orbital_functions import calc_jl
from cryspy.A_functions_base.database import DATABASE


def get_n_zeta_coeff_for_atom_with_orbital(atom_name: str, orbital_name: str):
    d_radial_slater = DATABASE["Orbitals by radial Slater functions"]
    n, zeta, coeff = d_radial_slater[(atom_name.capitalize(), orbital_name.lower())]
    return n, zeta, coeff

def get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol(ion_name: str):
    """
    Example: ion_name = "13C2-"
    """
    ion_name_sh = ion_name.strip().lower()
    flag_isotope, flag_charge = False, False
    isotope_number = ""
    charge_position = None

    for i_letter, letter in enumerate(ion_name_sh):
        if letter.isdigit():
            if i_letter == len(isotope_number):
                isotope_number += letter
                flag_isotope = True
            else:
                flag_charge = True
                charge_position = i_letter
                break

    if flag_isotope:
        isotope_position = len(isotope_number)
        isotope_number = int(isotope_number)
    else:
        isotope_position = 0
        isotope_number = None

    if flag_charge:
        atom_name = ion_name_sh[isotope_position:(charge_position+isotope_position)].capitalize()
        ion_charge = ion_name_sh[(charge_position+isotope_position):]
        ion_charge = int(f"{ion_charge[-1]:}{ion_charge[:-1]:}")
    else:
        atom_name = ion_name_sh[isotope_position:].capitalize()
        ion_charge = 0
    return atom_name, ion_charge, isotope_number

def get_atom_name_ion_charge_shell(ion_name: str):
    atom_name, ion_charge, isotop_number = get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol(ion_name)
    full_shell = get_full_shell(atom_name, ion_charge)
    return atom_name, ion_charge, full_shell

def add_ion_charge_to_shell(shell: str, ion_charge: int):
    shells = ["s", "p", "d", "f"]
    max_electrons = [2, 6, 10, 14]
    if ion_charge == 0:
        shell_out = shell
    else:
        l_orbital = shell.strip().split()
        if ion_charge > 0:
            l_orbital_new = []
            for orbital in reversed(l_orbital):
                n_shell, orbit_name, n_el = get_shell_orbit_electrons(orbital)
                if n_el > ion_charge:
                    orbital_new = f"{n_shell:}{orbit_name:}{n_el-ion_charge:}"
                    ion_charge = 0
                    l_orbital_new.append(orbital_new)
                elif n_el <= ion_charge:
                    ion_charge -= n_el
                elif ion_charge == 0:
                    l_orbital_new.append(orbital)
            l_orbital_new.reverse()
        else:
            l_orbital_new = []
            for orbital in l_orbital:
                n_shell, orbit_name, n_el = get_shell_orbit_electrons(orbital)
                n_max = max_electrons[shells.index(orbit_name)]
                n_add_max = n_max-n_el
                if n_add_max > 0:
                    if abs(ion_charge) <= n_add_max:
                        orbital_new = f"{n_shell:}{orbit_name:}{n_el-ion_charge:}"
                        ion_charge = 0
                    else:
                        orbital_new = f"{n_shell:}{orbit_name:}{n_max:}"
                        ion_charge += n_add_max
                    l_orbital_new.append(orbital_new)
                else:
                    l_orbital_new.append(orbital)
        shell_out = " ".join(l_orbital_new)
    if ion_charge != 0:
        raise UserWarning("Too much electrons for the valence shell")
    return shell_out


def get_shell_orbit_electrons(orbital:str):
    n_shell = int(orbital[0])
    orbit_name = orbital[1]
    n_el = int(orbital[2:])
    return n_shell, orbit_name, n_el


def get_full_shell(atom_name: str, ion_charge: int = 0):
    d_elements = DATABASE["Elements"]
    core_sh = d_elements["Orbital occupancy", atom_name]
    l_shell = core_sh.split("]")
    if len(l_shell) == 1:
        shell_full = l_shell[0].strip()
        shell_full = add_ion_charge_to_shell(shell_full, ion_charge)
    else:
        atom_name_core = l_shell[0][1:]
        shell_start_full = get_full_shell(atom_name_core)
        shell_end = l_shell[1].strip()
        shell_end = add_ion_charge_to_shell(shell_end, ion_charge)
        shell_full = shell_start_full + " " + shell_end
    return shell_full


def calc_jl_for_shell_of_ion(sthovl, ion_name: str, shell: str, kappa:float = 1., l_max:int=5):
    n_shell, orbit_name = shell[0], shell[1]
    atom_name, ion_charge, full_shell = get_atom_name_ion_charge_shell(ion_name)
    n_total_el = sum([get_shell_orbit_electrons(orbital)[2] for orbital in full_shell.split()])
    jl_res = None
    d_radial_slater = DATABASE["Orbitals by radial Slater functions"]
    for orbital in shell.strip().split():
        n_shell, orbit_name, n_el = get_shell_orbit_electrons(orbital)
        s_name = f"{atom_name.capitalize()}_{orbit_name.lower()}"
        shell = f"{n_shell:}{orbit_name:}"
        try:
            n, zeta, coeff = d_radial_slater[(atom_name.capitalize(), shell.lower())]
            # 
            # d_atom = D_ATOM_RHO_ORBITAL_RADIAL_SLATER[atom_name.capitalize(), orbit_name.lower()]
            # n = d_atom["n0"]
            # zeta = d_atom["zeta0"]
            # coeff = d_atom[shell]
            # if numpy.any(numpy.isnan(coeff)):
            #     raise KeyError
        except KeyError: #FIXME
            n = numpy.array([n_shell,], dtype=int)
            zeta = numpy.array([n_total_el,], dtype=float)
            coeff = numpy.array([1,], dtype=float)
            raise UserWarning(f"Parameters for '{s_name:}' are not found")

        jl = calc_jl(sthovl, coeff, n, zeta, kappa=kappa, l_max=l_max)
        if jl_res is None:
            jl_res = jl*n_el
        else:
            jl_res += jl*n_el
    return jl_res


def calc_jl_for_ion(sthovl, ion_name: str, kappa:float = 1., l_max:int=5):
    atom_name, ion_charge, full_shell = get_atom_name_ion_charge_shell(ion_name)
    jl_res = calc_jl_for_shell_of_ion(sthovl, ion_name, full_shell, kappa = kappa, l_max=l_max)
    return jl_res


def calc_scattering_amplitude_tabulated(ion_name: str, sthovl, flag_sthovl: bool = False):
    d_scattering_amplitude = DATABASE["Scattering amplitude"]
    params = d_scattering_amplitude[ion_name.capitalize()]
    a1, b1 = params[0], params[1]
    a2, b2 = params[2], params[3]
    a3, b3 = params[4], params[5]
    a4, b4 = params[6], params[7]
    c = params[8]
    
    sthovl_sq = numpy.square(sthovl)
    scattering_amplitude = (
        a1*numpy.exp(-b1*sthovl_sq) + 
        a2*numpy.exp(-b2*sthovl_sq) + 
        a3*numpy.exp(-b3*sthovl_sq) + 
        a4*numpy.exp(-b4*sthovl_sq) + 
        c
    )
    dder = {}
    if flag_sthovl:
        dder["sthovl"] = -2.*sthovl*(
            a1*b1*numpy.exp(-b1*sthovl_sq) + 
            a2*b2*numpy.exp(-b2*sthovl_sq) + 
            a3*b3*numpy.exp(-b3*sthovl_sq) + 
            a4*b4*numpy.exp(-b4*sthovl_sq)
        )
    return scattering_amplitude, dder



"""DensityPoint and DensityPointL classes."""
from typing import NoReturn
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

import numpy

class ChannelChi(ItemN):
    """
    Magnetization density in channel chi.

    Attributes
    ----------
        - numerator_x, numerator_y, numerator_z, denominator_xyz (mandatory)
        - index_x, index_y, index_z (optional)
        - density
        - chi_11, chi_22, chi_33, chi_12, chi_13, chi_23, chi_21, chi_31, chi_32 
        - atom_multiplicity
        - point_multiplicity (optional)
    """

    ATTR_MANDATORY_NAMES = ("numerator_x", "numerator_y", "numerator_z", "denominator_xyz",
        "density",
        "chi_11", "chi_22", "chi_33", "chi_12", "chi_13", "chi_23", "chi_21", "chi_31", "chi_32", 
        "atom_multiplicity", )
    ATTR_MANDATORY_TYPES = (int, int, int, int,
        float,
        float, float, float, float, float, float, float, float, float, 
        int)
    ATTR_MANDATORY_CIF = ("numerator_x", "numerator_y", "numerator_z", "denominator_xyz",
        "density",
        "chi_11", "chi_22", "chi_33", "chi_12", "chi_13", "chi_23", "chi_21", "chi_31", "chi_32", 
        "atom_multiplicity", )

    ATTR_OPTIONAL_NAMES = ("index_x", "index_y", "index_z", "point_multiplicity", )
    ATTR_OPTIONAL_TYPES = (int, int, int, int)
    ATTR_OPTIONAL_CIF = ("index_x", "index_y", "index_z", "point_multiplicity", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {
        "density": "{:.5f}",
        "chi_11": "{:.3f}", "chi_12": "{:.3f}", "chi_13": "{:.3f}", 
        "chi_21": "{:.3f}", "chi_22": "{:.3f}", "chi_23": "{:.3f}",
        "chi_31": "{:.3f}", "chi_32": "{:.3f}", "chi_33": "{:.3f}",}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "channel_chi"

    def __init__(self, **kwargs) -> NoReturn:
        super(ChannelChi, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"density": 0.,
            "point_multiplicity": 0, "atom_multiplicity": 0,}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class ChannelChiL(LoopN):
    """
    Magnetization density in channel chi in loop

    Attributes
    ----------
        - numerator_x, numerator_y, numerator_z, denominator_xyz (mandatory)
        - index_x, index_y, index_z (optional)
        - density_11, density_22, density_33, density_12, density_13, density_23, density_21, density_31, density_32 
        - atom_multiplicity
        - point_multiplicity (optional)
    """

    ITEM_CLASS = ChannelChi
    ATTR_INDEX = None

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(ChannelChiL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def get_dictionary(self):
        atom_multiplicity_channel_chi = numpy.array(self.atom_multiplicity, dtype=int)
        point_multiplicity_channel_chi = numpy.array(self.point_multiplicity, dtype=int)
        symm_elem_channel_chi = numpy.array([self.numerator_x, self.numerator_y, self.numerator_z, self.denominator_xyz], dtype=int)
        susceptibility_channel_chi = numpy.array([
            self.chi_11, self.chi_12, self.chi_13,
            self.chi_21, self.chi_22, self.chi_23,
            self.chi_31, self.chi_32, self.chi_33], dtype=float)
        density_channel_chi = numpy.array(self.density, dtype=float)

        dict_out = {
            "atom_multiplicity_channel_chi": atom_multiplicity_channel_chi,
            "point_multiplicity_channel_chi": point_multiplicity_channel_chi,
            "symm_elem_channel_chi": symm_elem_channel_chi,
            "susceptibility_channel_chi": susceptibility_channel_chi,
            "density_channel_chi": density_channel_chi,
            }
        return dict_out
